# =============================================================================
# Fog Simhash
# =============================================================================
#
# Classes & functions related to the Simhash Local Sensitivity Hashing scheme.
#
# [Urls]:
# https://en.wikipedia.org/wiki/SimHash
#
# [Reference]:
# Charikar, Moses S. (2002), "Similarity estimation techniques from rounding
# algorithms", Proceedings of the 34th Annual ACM Symposium on Theory of
# Computing.
#
import hashlib

from fog.lsh.utils import is_power_of_two, popcount

# TODO: reimplement soundly by following:
# https://github.com/reubano/changanya/blob/master/changanya/simhash.py
# Note: Simhash clustering methods I could find are two slow (mozilla's one)
# and even Hamming space indexation ones.
# I should try again to use the same lsh method used by minhash to see if it
# yields similar results

# TODO: reference papers
# https://github.com/leonsim/simhash/blob/master/simhash/__init__.py#L116-L208
# https://github.com/scrapinghub/python-simhash
# http://www.wwwconference.org/www2007/papers/paper215.pdf
# https://github.com/seomoz/simhash-cpp/tree/e7aacb1642f406ff0815cf402e909d2002473812
# https://ir.library.dc-uoit.ca/bitstream/10155/475/1/Rodriguez%20Reina_Ernesto.pdf


def simhash(tokens, f=128):
    assert f <= 512 and not is_power_of_two(f), 'fog.lsh.simhash: f should be a power of 2 <= 512'

    histogram = [0] * f

    l = f >> 2
    hash_func = hashlib.md5

    if f == 256:
        hash_func = hashlib.sha256

    elif f == 512:
        hash_func = hashlib.sha512

    for token in tokens:
        h = hash_func()
        h.update(token.encode())
        h = int(h.hexdigest()[0:l], 16)

        for i in range(f):
            r = ((h >> i) & 1)
            histogram[i] += r - (r ^ 1)

    signature = 0

    for i in range(f):
        if histogram[i] > 0:
            signature |= (1 << i)

    return signature


def simhash_distance(A, B, f=128):
    return popcount(A ^ B) / f


def simhash_similarity(A, B, f=128):
    return 1.0 - popcount(A ^ B) / f


if __name__ == '__main__':
    from fog.metrics import cosine_similarity
    from fog.tokenizers import ngrams

    TESTS = [
        (
            'the cat sat on the mat',
            'the cat sat on a mat'
        ),
        (
            'whatever floats your goat',
            'whatever floats your moat'
        ),
        (
            'hello darkness my old friend',
            'goodbye darkness my old friends'
        ),
        (
            'aaabbc',
            'zzzyyx'
        )
    ]

    for s1, s2 in TESTS:
        print(s1)
        print(s2)
        s1 = list(ngrams(2, s1))
        s2 = list(ngrams(2, s2))

        for i in [32, 64, 128, 256, 512]:
            sm1 = simhash(s1, f=i)
            sm2 = simhash(s2, f=i)

            print(i)
            print(sm1, sm2)
            print(cosine_similarity(s1, s2))
            print(simhash_similarity(sm1, sm2, f=i))
            print()



# =============================================================================
# Fog Skeleton Key
# =============================================================================
#
# The skeleton key by Pollock and Zamora.
#
# [Urls]:
# http://dl.acm.org/citation.cfm?id=358048
# http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.12.385&rep=rep1&type=pdf
#
# [Reference]:
# Pollock, Joseph J. and Antonio Zamora. 1984. "Automatic Spelling Correction
# in Scientific and Scholarly Text." Communications of the ACM, 27(4).
# 358--368.
import re
from unidecode import unidecode

UNDESIRABLES_RE = re.compile(r'[^A-Z]')
VOWELS = set('AEIOU')


def skeleton_key(string):
    """
    Function returning a string's skeleton key which is constructed thusly:

    1. The first letter of the string
    2. Unique consonants in order of appearance
    3. Unique vowels in order of appearance

    This key is very useful when searching for mispelled strings because
    if sorted using this key, similar strings will be next to each other.

    Args:
        string (str): The string to encode.

    Returns:
        string: The string's skeleton key.

    """

    # Deburring
    string = unidecode(string)

    # Normalizing case
    string = string.upper()

    # Dropping useless characters
    string = re.sub(UNDESIRABLES_RE, '', string)

    # Composing the key
    if not string:
        return ''

    first_letter = string[0]
    key = [first_letter]

    consonants = set()
    vowels = set()

    for i in range(1, len(string)):
        letter = string[i]

        if letter == first_letter:
            continue

        if letter in VOWELS:
            if letter not in vowels:
                vowels.add(letter)
                key.append(letter)
        else:
            if letter not in consonants:
                consonants.add(letter)
                key.insert(len(key) - len(vowels), letter)

    return ''.join(key)



# =============================================================================
# Fog Key Collision Clustering
# =============================================================================
#
# Clustering algorithm grouping items by key collision.
#
# [Url]:
# http://openrefine.org/
#
from collections import defaultdict

from fog.clustering.utils import clusters_from_buckets


def key_collision(data, key=None, keys=None, min_size=2, max_size=float('inf'),
                  merge=True):
    """
    Function returning an iterator over found clusters.

    It works by grouping items in buckets if they share the same key, as
    returned by a key function. Useful key functions that can efficiently
    regroup similar strings range from case normalization to phonetic encodings
    to clever fingerprint functions such as the one used by OpenRefine.

    Since it runs in O(n), it is one of the fastest record-linkage clustering
    algorithm. However, it won't be able to match more subtle string
    differences such as typos etc.

    Note that the key function can return more than one key for the string.
    In that case, buckets can be merged or not and produce slightly different
    clusters. When using multiple keys per item, if the key space is dense, you
    should definitely merge buckets or you will end up with a large number
    of very similar clusters. If key space is sparse, you can avoid merging
    as an optimization strategy.

    Note also that falsey key are dropped.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        key (callable): A function returning an item's key.
        keys (callable): A function returning an item's keys.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        merge (bool, optional): whether to merge the buckets to form clusters.
            Defaults to True.

    Yield:
        list: A viable cluster.

    """

    buckets = defaultdict(list)

    # Single key
    if key is not None:
        for item in data:
            k = key(item)

            if k:
                buckets[k].append(item)

    # Multiple keys
    elif keys is not None:
        for item in data:
            ks = keys(item)

            if not ks:
                continue

            for k in ks:
                if k:
                    buckets[k].append(item)

    # Merging clusters
    if merge and key is None:
        yield from clusters_from_buckets(buckets.values(), fuzzy=True)

    # Buckets as clusters
    else:
        for cluster in buckets.values():
            if len(cluster) < min_size or len(cluster) > max_size:
                continue
            yield cluster



# =============================================================================
# Fog Sorted Neighborhood Clustering
# =============================================================================
#
# Implementation of the Sorted Neighborhood method.
#
# [References]:
# Mauricio A. Hernandez and Salvatore J. Stolfo. The merge/purge problem for
# large databases. In Proceedings of the ACM International Conference on
# Management of Data (SIGMOD), 1995.
#
# Mauricio A. Hernandez and Salvatore J. Stolfo. Real-world data is dirty:
# Data cleansing and the merge/purge problem. Data Mining and Knowledge
# Discovery, 2(1), 1998
#
# Yan, Su, Dongwon Lee, Min-Yen Kan, et Lee C. Giles. « Adaptive Sorted
# Neighborhood Methods for Efficient Record Linkage », 185. ACM Press, 2007.
# https://doi.org/10.1145/1255175.1255213.
#
# [Urls]:
# https://hpi.de/fileadmin/user_upload/fachgebiete/naumann/folien/SS13/DPDC/DPDC_14_SNM.pdf
#
from fog.clustering.utils import make_similarity_function, clusters_from_pairs

# TODO: parallelize
# TODO: default zigzag
# TODO: multipass snm variant


def sorted_neighborhood(data, key=None, keys=None, similarity=None, distance=None,
                        radius=None, window=10, min_size=2, max_size=float('inf'),
                        mode='connected_components'):
    """
    Function returning an iterator over found clusters using the sorted
    neighborhood method (SNM).

    It works by first sorting the data according to a key which could, if
    cleverly chosen, put similar items next to one another in the result.

    We then attempt to find clusters by computing pairwise similarity/distances
    in small blocks of constant size in the sorted list.

    Omission key & skeleton keys by Pollock & Zamora are a good choice of
    sorting key if you try to find mispellings, for instance.

    Other good choice of keys is a combination of sorting strings
    lexicographically and sorting them reversed ([None, lambda x: x[::-1]]).

    Note that the sorted neighboorhood method usually runs faster than blocking
    but also misses much more true positives.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        key (callable, optional): key on which to sort the data.
        keys (iterable, optional): list of keys on which to sort for multipass
            sorted neighborhood method.
        similarity (callable): If radius is specified, a function returning
            the similarity between two points. Else, a function returning
            whether two points should be deemed similar. Alternatively, one can
            specify `distance` instead.
        distance (callable): If radius is specified, a function returning
            the distance between two points. Else, a function returning
            whether two point should not be deemed similar. Alternatively, one
            can specify `similarity` instead.
        radius (number, optional): produced clusters' radius.
        window (number, optional): Size of the window in which to look for
            matches. Defaults to 10.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        mode (string, optional): 'fuzzy_clusters', 'connected_components'.
            Defaults to 'connected_components'.

    Yields:
        list: A viable cluster.

    """

    # Formatting similarity
    similarity = make_similarity_function(similarity=similarity, distance=distance, radius=radius)

    # Iterating over sorted data
    def clustering():
        multipass_keys = keys if keys is not None else [key]

        for k in multipass_keys:
            S = sorted(data, key=k)
            n = len(S)

            for i in range(n):
                A = S[i]

                for j in range(i + 1, min(n, i + window)):
                    B = S[j]

                    if similarity(A, B):
                        yield (A, B)

    # Building clusters
    yield from clusters_from_pairs(
        clustering(),
        min_size=min_size,
        max_size=max_size,
        mode=mode,
        fuzzy=keys is not None
    )


def full_AA_SNM(sorted_records, similarity, window):
    n = len(sorted_records)
    w = window
    step = w - 1
    first = 0
    last = first + step

    while last < n:
        f = first

        # Enlargement
        while similarity(sorted_records[first], sorted_records[last]):
            w = last - f + 1
            first = last
            last = first + w - 1

        # Retrenchment
        w = last - f + 1

        while w > window:
            previous = last - 1

            if similarity(sorted_records[previous], sorted_records[last]):
                break

            # NOTE: possibility to divide the window by two according to
            # the paper.
            w -= 1
            last = previous

        w = window

        yield (f, last)

        first = last + 1
        last = first + step

    if first < n:
        yield (first, n - 1)


def adaptive_sorted_neighborhood(data, key=None, keys=None, similarity=None,
                                 distance=None, radius=None, window=10,
                                 min_size=2, max_size=float('inf'),
                                 mode='connected_components'):
    """
    Function returning an iterator over found clusters using the adaptive
    sorted neighborhood method (ASNM).

    More specifically, this method implements the "Full-Accumulatively-Adaptative
    SNM" from the "Adaptive Sorted Neighborhood Methods for Efficient
    Record Linkage" paper.

    Note that this method is more precise that basic SNM only for cases where
    duplicate components are sufficiently large and when the variance
    of their size is important. Else it's basically useless.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        key (callable, optional): key on which to sort the data.
        keys (iterable, optional): list of keys on which to sort for multipass
            sorted neighborhood method.
        similarity (callable): If radius is specified, a function returning
            the similarity between two points. Else, a function returning
            whether two points should be deemed similar. Alternatively, one can
            specify `distance` instead.
        distance (callable): If radius is specified, a function returning
            the distance between two points. Else, a function returning
            whether two point should not be deemed similar. Alternatively, one
            can specify `similarity` instead.
        radius (number, optional): produced clusters' radius.
        window (number, optional): Size of the window in which to look for
            matches. Defaults to 10.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        mode (string, optional): 'fuzzy_clusters', 'connected_components'.
            Defaults to 'connected_components'.

    Yields:
        list: A viable cluster.

    """

    # Formatting similarity
    similarity = make_similarity_function(similarity=similarity, distance=distance, radius=radius)

    # Iterating over sorted data
    def clustering():
        multipass_keys = keys if keys is not None else [key]

        for k in multipass_keys:
            S = sorted(data, key=k)

            for start, end in full_AA_SNM(S, similarity, window):
                l = end - start + 1

                if l == 1:
                    continue

                elif l == 2:
                    A = S[start]
                    B = S[end]

                    if similarity(A, B):
                        yield (A, B)

                else:
                    for i in range(start, end + 1):
                        for j in range(i + 1, end + 1):
                            A = S[i]
                            B = S[j]

                            if similarity(A, B):
                                yield (A, B)

    # Building clusters
    yield from clusters_from_pairs(
        clustering(),
        min_size=min_size,
        max_size=max_size,
        mode=mode,
        fuzzy=keys is not None
    )



# =============================================================================
# Fog Ngrams Tokenizers
# =============================================================================
#
# Functions related to n-grams.
#
from functools import partial


# TODO: add positional n-grams
def ngrams(n, tokens):
    """
    Function returning an iterator over the given sequence's n-grams. The
    type of yielded n-grams will match the sequence's one.

    Args:
        n (number): size of the subsequences.
        tokens (iterable): token sequence, should support random access.

    Yields:
        tuple or string: A n-gram.

    """
    lt = len(tokens)

    if lt < n:
        yield tokens

    limit = lt - n + 1

    for i in range(limit):
        yield tokens[i:i + n]


bigrams = partial(ngrams, 2)
trigrams = partial(ngrams, 3)
quadrigrams = partial(ngrams, 4)


def join_grams_iter(grams):
    yield next(grams)

    for g in grams:
        yield g[-1]


def join_ngrams(grams):
    return ''.join(join_grams_iter(grams))



# =============================================================================
# Fog Evaluation Utils
# =============================================================================
#
# Miscellaneous helper functions related to evaluation.
#
from sys import version_info
from collections import OrderedDict
from collections.abc import Mapping, Sequence

AT_LEAST_PY37 = version_info >= (3, 7)
ordered_dict = dict if AT_LEAST_PY37 else OrderedDict


def labels_to_clusters(labels):
    clusters = ordered_dict()

    iterator = labels

    if isinstance(labels, Mapping):
        iterator = labels.items()
    elif isinstance(labels, Sequence):
        iterator = enumerate(labels)

    for i, l in iterator:
        c = clusters.get(l)

        if c is None:
            c = []
            clusters[l] = c

        c.append(i)

    return list(clusters.values())


def clusters_to_labels(clusters, *, key=None, flat=False):
    labels = {}

    for i, cluster in enumerate(clusters):
        for item in cluster:
            k = item

            if key is not None:
                k = key(item)

            labels[k] = i

    if flat:
        flat_labels = [-1] * len(labels)

        for i, l in labels.items():
            flat_labels[i] = l

        return flat_labels

    return labels



# =============================================================================
# Fog Rusalka
# =============================================================================
#
# Rusalka is a standardization algorithm for transliterated Russian names.
#
import re
from unidecode import unidecode

# Notes
# Version renvoyant des nombres
# skaïa skaia skaja
# Doïstoievski
# Alexeïevitch Alexeievitch
# Shchedrin Ščedrin Schtschedrin Chtchedrine Szczedrin Sxedrín Sjtjedrin Scsedrin Sjtsjedrin Șcedrin Štšedrin
# Yevgeniy Eugene
# Pjotr
# Tatiana Tatjana
# Feodor Fyodor Fjodor
# Růžena Maturová
# Hranislav
# Sergeiy_Liminovich
# q -> k
# Almas_Uteschow, Almas_Uteshov
# Dmytro_Zamotayev, Dmytro_Samotajew
# Farid_Dosayev, Farid_Dossajew
# Mirian_Giorgadze, Miriani_Giorgadse
# Musa_Murtazaliev, Musa_Murtasalijew
# Soslan_Fraev, Soslan_Frajew
# Volodymyr_Veredyuk, Wolodymyr_Weredjuk
# Roman_Kolzow, Roman_Koltsov
# Sergei_Belych, Sergey_Belykh
# wojciek
# Djezira, Giacomo, Djacomo al- el-
# Strzoda
# Variantes polonaises
# Pierrick_Rakotoharisoa,Pierrick_Rakotoharizoa
# Jughashvilli

# criblage à base de aev pour normalisers

# fails if given numbers

# j between vowels y

# http://intranet.library.arizona.edu/users/brewerm/sil/lib/transprimer.html
# http://intranet.library.arizona.edu/users/brewerm/sil/lib/transcription.html
# http://intranet.library.arizona.edu/users/brewerm/sil/lib/transbib.html

# http://usefulenglish.ru/vocabulary/russian-names-in-english-en

# â -> ja
# Khokhloma -> hohloma

FILTER = re.compile(r'[^a-z]')
DUPLICATED = re.compile(r'(.)\1+')
VOWELS = re.compile(r'[aeiouy]')

PHONETIC_NOTATION_RULES = [
    (r'[čšŝ]', 'ch'),
    (r'û', 'iou'),
    (r'ž', 'zh')
]

RULES = [
    (r'cz', 'ts'),
    (r'i[jy]$', ''),
    (r'[fvw]|ph', 'f'),
    (r'^[jy]?e', 'J'),
    (r'[dt]?[cs]h|tz', 'ʃ'),
    (r'[dt]?(?:zh|j)', 'ʒ'),
    (r'ks', 'x'),
    (r'^y(?=[aeiou])|iou', 'j'),
    (r'[cg]', 'k'),
    (r'h', '')
]

RULES = [(re.compile(p), r) for p, r in RULES]


def rusalka(name):

    # Lower case
    name = name.lower()

    # Applying phonetic notation rules
    for pattern, replacement in PHONETIC_NOTATION_RULES:
        name = re.sub(pattern, replacement, name)

    # Unidecode
    name = unidecode(name)

    # Dropping irrelevant characters
    name = re.sub(FILTER, '', name)

    if len(name) < 2:
        return name.lower()

    # Applying rules
    for pattern, replacement in RULES:
        name = re.sub(pattern, replacement, name)

    # Deduplication
    name = re.sub(DUPLICATED, r'\1', name)
    name = name.lower()

    first_letter = name[0]
    rest = name[1:]

    # Dropping vowels
    rest = re.sub(VOWELS, '', rest)

    # Handling first letter
    if re.match(VOWELS, first_letter):
        first_letter = 'a'

    return first_letter + rest



# =============================================================================
# Fog Cologne Phonetic Algorithm
# =============================================================================
#
# Function computing the Cologne phonetic code for German names.
#
# [Url]:
# https://en.wikipedia.org/wiki/Cologne_phonetics
#
# [Article]:
# Hans Joachim Postel: Die Kölner Phonetik. Ein Verfahren zur Identifizierung
# von Personennamen auf der Grundlage der Gestaltanalyse.
# in: IBM-Nachrichten, 19. Jahrgang, 1969, S. 925-931.
#
import re
from unidecode import unidecode
from fog.utils import squeeze

ALPHA_RE = re.compile(r'[^A-Z]')

CODES = {
    'H': None,

    'A': '0',
    'E': '0',
    'I': '0',
    'O': '0',
    'U': '0',
    'J': '0',
    'Y': '0',

    'B': '1',
    'P': '1',

    'F': '3',
    'V': '3',
    'W': '3',

    'G': '4',
    'K': '4',
    'Q': '4',

    'L': '5',

    'M': '6',
    'N': '6',

    'R': '7',

    'S': '8',
    'Z': '8'
}

DT = set(['C', 'S', 'Z'])
CFOLLOWING1 = set(['A', 'H', 'K', 'L', 'O', 'Q', 'R', 'U', 'X'])
CFOLLOWING2 = set(['A', 'H', 'K', 'O', 'Q', 'U', 'X'])
CPREVIOUS = set(['S', 'Z'])
X = set(['C', 'Q', 'K'])

GERMANIC_SUBSTITUTIONS = [
    ('Ä', 'A'),
    ('Ö', 'O'),
    ('Ü', 'U'),
    ('ß', 'SS'),
    ('PH', 'F')
]


def cologne(name):
    """
    Function returning the Cologne phonetic code for the given name.

    Args:
        name (string): Name to encode.

    Returns:
        string: The name encoded.

    """

    # Preparing the name
    name = name.upper()

    for p, r in GERMANIC_SUBSTITUTIONS:
        name = name.replace(p, r)

    name = unidecode(name)
    name = re.sub(ALPHA_RE, '', name)

    code = []

    last_i = len(name) - 1

    for i, letter in enumerate(name):
        possible_code = CODES.get(letter)

        if possible_code is not None:
            code.append(possible_code)

        # Handling D/T
        elif letter == 'D' or letter == 'T':
            code.append('8' if i < last_i and name[i + 1] in DT else '2')

        # Handling C
        elif letter == 'C':

            if (
                len(name) > 1 and
                (i == 0 and name[i + 1] in CFOLLOWING1)
                or (i < last_i and name[i + 1] in CFOLLOWING2 and name[i - 1] not in CPREVIOUS)
            ):
                code.append('4')
            else:
                code.append('8')

        # Handling X
        elif letter == 'X':
            code.append('8' if i > 0 and name[i - 1] in X else '48')

    if len(code) == 0:
        return ''

    # Squeezing and dropping not leading 0
    rest = squeeze(''.join(code[1:])).replace('0', '')

    return code[0] + rest



#!/usr/bin/env python
# =============================================================================
# Fog CLI Endpoint
# =============================================================================
#
# CLI Interface of the Fog library.
#
import sys
from argparse import ArgumentParser, FileType

from fog.cli.cluster import cluster_action, CLUSTERING_ROUTINES
from fog.cli.split import split_action
from fog.cli.transform import transform_action, OPERATIONS

SUBPARSERS = {}


def main():
    parser = ArgumentParser(prog='fog')
    subparsers = parser.add_subparsers(help='action to execute', title='actions', dest='action')

    cluster_subparser = subparsers.add_parser('cluster', description='Fuzzy clustering on a CSV column.')
    cluster_subparser.add_argument('column', help='column')
    cluster_subparser.add_argument('file', help='csv file to cluster', type=FileType('r'), default=sys.stdin, nargs='?')
    cluster_subparser.add_argument('-o', '--output', help='output file', type=FileType('w'), default=sys.stdout)
    cluster_subparser.add_argument('-a', '--algorithm', help='algorithm to use', default='fingerprint_collision', choices=list(CLUSTERING_ROUTINES.keys()))
    cluster_subparser.add_argument('-t', '--type', help='reporting file type', default='toml', choices=['html', 'toml'])
    SUBPARSERS['cluster'] = cluster_subparser

    split_subparser = subparsers.add_parser('split', description='Split a multivalued column into different lines.')
    split_subparser.add_argument('column', help='column')
    split_subparser.add_argument('file', help='csv file to split', type=FileType('r'), default=sys.stdin, nargs='?')
    split_subparser.add_argument('-o', '--output', help='output file', type=FileType('w'), default=sys.stdout)
    split_subparser.add_argument('-s', '--separator', help='values separator. Defaults to "|".', default='|')
    split_subparser.add_argument('-t', '--target-column', help='name of the column to create')
    split_subparser.add_argument('-r', '--rename-column', help='new name for the column')
    SUBPARSERS['split'] = split_subparser

    transform_subparser = subparsers.add_parser('transform', description='Transform the values of a column in batch.')
    transform_subparser.add_argument('column', help='column')
    transform_subparser.add_argument('operations', help='operations to apply, separated by commas {%s}' % ','.join(list(OPERATIONS.keys())))
    transform_subparser.add_argument('file', help='csv file to transform', type=FileType('r'), default=sys.stdin, nargs='?')
    transform_subparser.add_argument('-o', '--output', help='output file', type=FileType('w'), default=sys.stdout)
    transform_subparser.add_argument('-t', '--target-column', help='name of the column to create')
    transform_subparser.add_argument('-a', '--after', help='whether to add the new column just after the original one', action='store_true')
    transform_subparser.add_argument('--eval', help='evaluate the operations as a python expression rather than using an operation chain', action='store_true')
    SUBPARSERS['transform'] = transform_subparser

    help_suparser = subparsers.add_parser('help')
    help_suparser.add_argument('subcommand', help='name of the subcommand')
    SUBPARSERS['help'] = help_suparser

    args = parser.parse_args()

    if args.action == 'help':
        target_subparser = SUBPARSERS.get(args.subcommand)

        if target_subparser is None:
            parser.print_help()
        else:
            target_subparser.print_help()

    if args.action == 'cluster':
        cluster_action(args)

    if args.action == 'split':
        split_action(args)

    if args.action == 'transform':
        transform_action(args)

    if args.action is None:
        parser.print_help()


if __name__ == '__main__':
    main()



import fog.clustering as clustering
import fog.graph as graph
import fog.join as join
import fog.key as key
import fog.lsh as lsh
import fog.metrics as metrics
import fog.phonetics as phonetics
import fog.tokenizers as tokenizers



# =============================================================================
# Fog NN-Descent Clustering
# =============================================================================
#
# Implementation of the probabilistic NN-Descent algorithm able to build
# an approximate k-nn graph from a dataset in subquadratic time.
#
# [Reference]:
# Dong, Wei, Charikar Moses, et Kai Li. « Efficient K-Nearest Neighbor Graph
# Construction for Generic Similarity Measures », 577. ACM Press, 2011.
# https://doi.org/10.1145/1963405.1963487.
#
# [Notes]:
# Recall is a function of the k parameter. k is also intimately related
# to the data's intrisic's dimensionality. One should increase k to increase
# recall, at the cost of poorer performance. Chosing log2(n) as k seems to
# be a generally good compromise.
#
import math
import random
from heapq import heapify, heapreplace
from fog.clustering.utils import clusters_from_pairs

# TODO: parallelize nn_descent_full


def sample(rng, N, k, i):
    """
    Function sampling k indices from the range 0-N without the i index.

    """

    S = set()

    while len(S) < k:
        random_index = rng.randint(0, N - 1)

        if random_index == i:
            continue

        S.add(random_index)

    return list(S)


def reverse(B):
    """
    Returns the list of in-neighbors from the list of out-neighbors.

    """
    R = [[] for _ in range(len(B))]

    for i, neighbors in enumerate(B):
        for _, j in neighbors:
            R[j].append(i)

    return R


def nn_descent(data, radius, similarity=None, distance=None, k=None,
               min_size=2, max_size=float('inf'),
               mode='connected_components',
               seed=None):
    """
    Function returning an iterator over found clusters using the NN-Descent
    algorithm.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        radius (number): produced clusters' radius.
        k (number, optional): number of nearest neighbor to find per item.
            If not given, k will default to log2(n).
        similarity (callable): If radius is specified, a function returning
            the similarity between two points. Else, a function returning
            whether two points should be deemed similar. Alternatively, one can
            specify `distance` instead.
        distance (callable): If radius is specified, a function returning
            the distance between two points. Else, a function returning
            whether two point should not be deemed similar. Alternatively, one
            can specify `similarity` instead.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        mode (string, optional): 'fuzzy_clusters', 'connected_components'.
            Defaults to 'connected_components'.
        seed (number, optional): Seed for RNG. Defaults to None.

    Yields:
        list: A viable cluster.

    """

    # Seeding rng
    rng = random.Random(seed)

    # Inverting distance if needed
    if distance is not None:
        similarity = lambda x, y: -distance(x, y)
        radius = -radius

    # Making data set into indexable list
    if type(data) is not list:
        data = list(data)

    # Note that B & R could be flat arrays
    V = data
    B = []
    N = len(V)

    # Chosing k
    if k is None:
        k = int(math.log2(N))

    # Initial samples
    for i, item in enumerate(V):
        neighbors = [(similarity(item, V[j]), j) for j in sample(rng, N, k, i)]
        heapify(neighbors)
        B.append(neighbors)

    c = 1

    while c != 0:
        R = reverse(B)
        C = []

        c = 0

        for i, item in enumerate(V):
            candidates = set(j for _, j in B[i])
            candidates.update(R[i])

            C.append(list(candidates))

        for i in range(N):
            BA = C[i]

            for ii in BA:
                BB = C[ii]

                for jj in BB:

                    if i == jj:
                        continue

                    s = similarity(V[i], V[jj])

                    if s > B[i][0][0]:
                        c += 1
                        heapreplace(B[i], (s, jj))

    def clustering():
        for i, neighbors in enumerate(B):
            for s, j in neighbors:
                if s >= radius:
                    yield (i, j)

    gen = clusters_from_pairs(
        clustering(),
        min_size=min_size,
        max_size=max_size,
        mode=mode,
        fuzzy=True
    )

    for cluster in gen:
        yield [V[i] for i in cluster]


def shuffle_k(rng, L, k):
    i = -1
    last_i = len(L) - 1

    while i < k - 1:
        i += 1

        r = rng.randint(i, last_i)
        v = L[r]

        L[r] = L[i]
        L[i] = v


def sample_full(rng, L, k):
    """
    Function sampling k indices from given list.

    """

    N = len(L)

    if N <= k:
        return list(L)

    shuffle_k(rng, L, k)

    return L[0:k]


def reverse_full(old, new):
    """
    Returns the list of in-neighbors from the list of out-neighbors.

    """
    old_prime = [[] for _ in range(len(old))]
    new_prime = [[] for _ in range(len(new))]

    for i in range(len(old)):
        for j in old[i]:
            old_prime[j].append(i)
        for j in new[i]:
            new_prime[j].append(i)

    return old_prime, new_prime


def nn_descent_full(data, radius, similarity=None, distance=None, k=None,
                    min_size=2, max_size=float('inf'),
                    mode='connected_components',
                    seed=None, rho=0.5, delta=0.001):
    """
    Function returning an iterator over found clusters using the NN-Descent Full
    algorithm.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        radius (number): produced clusters' radius.
        k (number, optional): number of nearest neighbor to find per item.
            If not given, k will default to log2(n).
        similarity (callable): If radius is specified, a function returning
            the similarity between two points. Else, a function returning
            whether two points should be deemed similar. Alternatively, one can
            specify `distance` instead.
        distance (callable): If radius is specified, a function returning
            the distance between two points. Else, a function returning
            whether two point should not be deemed similar. Alternatively, one
            can specify `similarity` instead.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        mode (string, optional): 'fuzzy_clusters', 'connected_components'.
            Defaults to 'connected_components'.
        seed (number, optional): Seed for RNG. Defaults to None.
        rho (number, optional): rho parameter, or sample rate, defining a
            trade-off between accuracy and speed. The lower this number is,
            the faster it runs, but the less accurate it is. Defaults
            to 0.5.
        delta (number, optional): early termination threshold. Roughly the
            fraction of try K-NN that are allowed to be missed due an early
            termination. Defaults to 0.001.

    Yields:
        list: A viable cluster.

    """

    # Seeding rng
    rng = random.Random(seed)

    # Inverting distance if needed
    if distance is not None:
        similarity = lambda x, y: -distance(x, y)
        radius = -radius

    # Making data set into indexable list
    if type(data) is not list:
        data = list(data)

    # Note that B & R could be flat arrays
    V = data
    B = []
    N = len(V)

    # Chosing k
    if k is None:
        k = int(math.log2(N))

    NK = N * k
    rhoK = int(rho * k)

    # Initial samples
    for i, item in enumerate(V):
        neighbors = [(similarity(item, V[j]), j, True) for j in sample(rng, N, k, i)]
        heapify(neighbors)
        B.append(neighbors)

    c = NK

    while c >= delta * NK:
        old = []
        new = []

        for i, item in enumerate(V):
            o = set()
            n = []

            for j, neighbor in enumerate(B[i]):
                if not neighbor[2]:
                    o.add(neighbor[1])
                else:
                    n.append((j, neighbor[1]))

            # Sampling rhoK from new
            s = sample_full(rng, n, rhoK)
            sample_set = set(s)

            for k in n:
                if k not in sample_set:
                    B[i][k[0]] = B[i][k[0]][0:2] + (False, )

            old.append(o)
            new.append(set([index for _, index in s]))

        old_prime, new_prime = reverse_full(old, new)

        c = 0

        for i, item in enumerate(V):
            o = old[i]
            n = new[i]

            o.update(sample_full(rng, old_prime[i], rhoK))
            n.update(sample_full(rng, new_prime[i], rhoK))

            for u1 in n:

                for u2 in n:
                    if u1 >= u2:
                        continue

                    s = similarity(V[u1], V[u2])

                    if s > B[u1][0][0]:
                        c += 1
                        heapreplace(B[u1], (s, u2, True))

                    if s > B[u2][0][0]:
                        c += 1
                        heapreplace(B[u2], (s, u1, True))

                for u2 in o:

                    if u1 == u2:
                        continue

                    s = similarity(V[u1], V[u2])

                    if s > B[u1][0][0]:
                        c += 1
                        heapreplace(B[u1], (s, u2, True))

                    if s > B[u2][0][0]:
                        c += 1
                        heapreplace(B[u2], (s, u1, True))

    def clustering():
        for i, neighbors in enumerate(B):
            for s, j, _ in neighbors:
                if s >= radius:
                    yield (i, j)

    gen = clusters_from_pairs(
        clustering(),
        min_size=min_size,
        max_size=max_size,
        mode=mode,
        fuzzy=True
    )

    for cluster in gen:
        yield [V[i] for i in cluster]



# =============================================================================
# Fog Key Intersection Clustering
# =============================================================================
#
# Clustering routine leveraging an intersection index to find similar items.
#
from collections import defaultdict, Counter
from fog.clustering.utils import clusters_from_pairs

# TODO: online tf-idf cutoff by automatically cutting down large buckets over
# a given threshold. call it burst and erase already set counters

# TODO: cosine variants

# NOTE: compromise between key collision & meta blocking

# TODO: found in the litterature as pair-count / probe count
# ref in sarawagi-kirpal

METRICS = ['jaccard', 'overlap']


def intersection_index(data, metric='jaccard', radius=0.8, key=None, min_size=2,
                       max_size=float('inf'), mode='connected_components'):
    """
    Function returning an iterator over found clusters.

    It works by building an index of the given documents' tokens' intersection.

    It runs in subquadratic time: the less probable it is for two tokens to
    collide, the more it will approach linear time. As such, this algorithm
    is a good fit when handling large ngrams, for instance, but will be very
    poor when considering single characters as tokens since it means
    document will collide often because they will naturally share a lot of
    letters.

    As such, note that it is possible to make this method faster by leveraging
    Zipf's law in some cases.

    Note also that this algorithm is on par with pairwise methods in that it
    does not try to approximate and will return all the existing pairs.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        metric (string, optional): Metric to use. One of {'jaccard', 'overlap'}.
            Defaults to 'jaccard'.
        radius (number, optional): Jaccard similarity radius.
        key (callable, optional): Function returning an item's key.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        mode (string, optional): 'fuzzy_clusters', 'connected_components'.
            Defaults to 'connected_components'.

    Yield:
        list: A viable cluster.

    """

    if metric not in METRICS:
        raise TypeError('fog.clustering.intersection_index: unknown metric "%s"' % metric)

    buckets = defaultdict(list)
    intersections = defaultdict(Counter)

    if type(data) is not list:
        data = list(data)

    sizes = [0] * len(data)

    for i, item in enumerate(data):
        if key is not None:
            item = key(item)

        shingles = set(item)
        sizes[i] = len(shingles)

        for shingle in shingles:
            bucket = buckets[shingle]

            for j in bucket:
                intersections[i][j] += 1

            bucket.append(i)

    def clustering():
        for i, neighbors in intersections.items():

            for j, I in neighbors.items():

                if metric == 'jaccard':
                    U = sizes[i] + sizes[j] - I

                    if I / U >= radius:
                        yield (i, j)

                else:
                    M = min(sizes[i], sizes[j])

                    if I / M >= radius:
                        yield (i, j)

    gen = clusters_from_pairs(
        clustering(),
        min_size=min_size,
        max_size=max_size,
        mode=mode
    )

    for cluster in gen:
        yield [data[i] for i in cluster]



# =============================================================================
# Fog Word Tokenizer
# =============================================================================
#
# A general purpose word tokenizer able to consider a lot of edge case and
# typical entities all while remaining mostly language agnostic wrt languages
# separating their words using whitespace (not the asian languages, for instance).
#
# It was mostly designed for French and English, but it probably works with
# other latin languages out of the box.
#
# The emitted tokens are tagged by entity types (not part-of-speech).
#
# Some design choices:
#  * We chose to only tag as numbers strings that could be parsed as ints or floats
#    without ambiguity. This means word tokens may contain things that
#    could be considered as numbers, but you can analyze them further down the line.
#
# Here is a list of things we don't handle (yet):
#   * Complex graphemes such as: u̲n̲d̲e̲r̲l̲i̲n̲e̲d̲ or ārrive
#   * Multi-line hyphenation schemes
#   * Junk found in the middle of a word token
#   * It is not possible to keep apostrophes starting names
#
import re
from html import unescape
from emoji import get_emoji_regexp
from unidecode import unidecode_expect_ascii
from ebbe import with_next
from typing import Optional, Iterable

DECIMALS = ['.', ',']
IDENTIFIER_PARTS = ['-', '_']
APOSTROPHES = ['\'', '’']
SIGILS = ['#', '@', '$']
IRISH = ['O', 'o']
VOWELS_PATTERN = 'aáàâäąåoôóøeéèëêęiíïîıuúùûüyÿæœAÁÀÂÄĄÅOÓÔØEÉÈËÊĘIİÍÏÎYŸUÚÙÛÜÆŒ'
CONSONANTS_RE = re.compile(r'^[^%s]$' % VOWELS_PATTERN)
EMAIL_LOOKAHEAD_RE = re.compile(r'^[A-Za-z0-9!#$%&*+\-/=?^_`{|}~]{1,64}@')
SMILEY_RE = re.compile(r'^(?:[\-]+>|<[\-]+|[<>]?[:;=8][\-o\*\']?[\)\]\(\[dDpP/\:\}\{@\|\\]|[\)\]\(\[dDpP/\:\}\{@\|\\][\-o\*\']?[:;=8]|[<:]3|\^\^)')
EMOJI_RE = get_emoji_regexp()
POINT_SPLITTER_RE = re.compile(r'(\.)')
LENGTHENING_RE = re.compile(r'([^\W\d])\1{4,}')

ENGLISH_CONTRACTIONS = ['ll', 're', 'm', 's', 've', 'd']
ENGLISH_ARCHAIC_CONTRACTIONS = ['tis', 'twas']
FRENCH_EXCEPTIONS = ['hui']
FRENCH_UNION_TARGETS = {
    'je',
    'tu',
    'il',
    'ils',
    'elle',
    'elles',
    'nous',
    'vous',
    'ils',
    'on',
    'le',
    'la',
    'les',
    'moi',
    'toi',
    'lui',
    'y'
}
ABBREVIATIONS = {
    'apt',
    'appt',
    'dr',
    'etc',
    'jr',
    'm',
    'mgr',
    'min',
    'mr',
    'mrs',
    'ms',
    'mme',
    'mlle',
    'no',
    'p',
    'pp',
    'prof',
    'sr',
    'st',
    'vs'
}

TOKEN_TYPES = {
    'word',
    'number',
    'punct',
    'emoji',
    'smiley',
    'url',
    'email',
    'hashtag',
    'mention'
}


def is_ascii_junk(c):
    return c <= '\x1f'


def is_valid_twitter_char(c):
    return (
        c == '_' or
        (c >= '0' and c <= '9') or
        (c >= 'a' and c <= 'z') or
        (c >= 'A' and c <= 'Z')
    )


def is_consonant(c):
    return bool(CONSONANTS_RE.match(c))


def starts_as_url(string):
    return string.startswith('http://') or string.startswith('https://')


def email_lookahead(string):
    return bool(EMAIL_LOOKAHEAD_RE.match(string))


def with_holes(gen, string):
    last_index = 0

    for m in gen:
        if m.start() > last_index:
            for i in range(last_index, m.start()):
                yield ('punct', string[i])

        yield ('emoji', m.group(0))
        last_index = m.end()

    if last_index < len(string):
        for i in range(last_index, len(string)):
            yield ('punct', string[i])


def punct_emoji_iter(string):
    skip_next = False

    for item, next_item in with_next(with_holes(EMOJI_RE.finditer(string), string)):
        if skip_next:
            skip_next = False
            continue

        # Emoji combinator
        if (
            next_item is not None and
            item[0] == 'emoji' and
            next_item[0] == 'punct' and
            next_item[1] == '\ufe0f'
        ):
            skip_next = True
            yield ('emoji', item[1] + next_item[1])

        else:
            yield item


def reduce_lenghtening(string):
    return LENGTHENING_RE.sub(r'\1\1\1', string)


def validate_token_types(types):
    for token_type in types:
        if token_type not in TOKEN_TYPES:
            raise TypeError('unknown token type: %s' % token_type)


def split_hashtag(string):
    offset = 1
    it = enumerate(string)
    state = 'upper-start' if not string[1].isdigit() else 'number'

    next(it)  # Hastag sigil '#'
    next(it)  # First char

    # NOTE: I hate state machines (I don't use a regex here to keep the benefit
    # of python's string methods wrt diacritics)
    for i, c in it:
        if state == 'lower':
            if c.isupper():
                yield ('word', string[offset:i])
                state = 'upper-start'
                offset = i

            elif c.isdigit():
                yield ('word', string[offset:i])
                state = 'number'
                offset = i

        elif state == 'upper-start':
            if c.islower():
                state = 'lower'
                continue

            if c.isupper():
                state = 'upper-next'
                continue

            if c.isdigit():
                state = 'number'
                yield ('word', string[offset:i])
                offset = i

        elif state == 'upper-next':
            if c.islower():
                state = 'lower'
                yield ('word', string[offset:i - 1])
                offset = i - 1

            elif c.isdigit():
                state = 'number'
                yield ('word', string[offset:i])
                offset = i

        elif state == 'number':
            if not c.isdigit():
                if c.isupper():
                    state = 'upper-start'
                    yield ('number', string[offset:i])
                    offset = i

                else:
                    state = 'lower'
                    yield ('number', string[offset:i])
                    offset = i

    yield ('number' if state == 'number' else 'word', string[offset:])


class WordTokenizer(object):
    def __init__(
        self,
        lower: bool = False,
        unidecode: bool = False,
        reduce_words: bool = False,
        decode_html_entities: bool = False,
        normalize_mentions: bool = False,
        normalize_hashtags: bool = False,
        mentions_as_words: bool = False,
        hashtags_as_words: bool = False,
        split_hashtags: bool = False,
        min_word_length: Optional[int] = None,
        max_word_length: Optional[int] = None,
        stoplist: Optional[Iterable[str]] = None,
        keep: Optional[Iterable[str]] = None,
        drop: Optional[Iterable[str]] = None
    ):
        self.lower = lower
        self.unidecode = unidecode
        self.reduce_words = reduce_words
        self.decode_html_entities = decode_html_entities
        self.normalize_mentions = normalize_mentions
        self.normalize_hashtags = normalize_hashtags
        self.mentions_as_words = mentions_as_words
        self.hashtags_as_words = hashtags_as_words
        self.split_hashtags = split_hashtags
        self.min_word_length = min_word_length
        self.max_word_length = max_word_length
        self.stoplist = stoplist
        self.keep = keep
        self.drop = drop

        if self.keep is not None and self.drop is not None:
            raise TypeError('giving both `keep` and `drop` makes no sense')

        if self.min_word_length is not None and self.max_word_length is not None:
            if self.min_word_length >= self.max_word_length:
                raise TypeError('`min_word_length` should be less than `max_word_length`')

        if self.stoplist is not None:
            if not isinstance(self.stoplist, set):
                self.stoplist = set(self.stoplist)

            if self.lower:
                self.stoplist = set(token.lower() for token in self.stoplist)

        if self.keep:
            self.keep = set(self.keep)
            validate_token_types(self.keep)

        if self.drop:
            self.drop = set(self.drop)
            validate_token_types(self.drop)

        self.__only_defaults = True

        if (
            self.lower or
            self.normalize_mentions or
            self.normalize_hashtags or
            self.mentions_as_words or
            self.hashtags_as_words or
            self.min_word_length is not None or
            self.max_word_length is not None or
            self.stoplist is not None or
            self.keep is not None or
            self.drop is not None
        ):
            self.__only_defaults = False

    def __tokenize(self, string):
        i = 0
        l = len(string)

        while i < l:
            c = string[i]

            # Chomping spaces and ASCII junk
            if c.isspace() or is_ascii_junk(c):
                i += 1
                continue

            can_be_mention = c == '@'
            can_be_hashtag = (c == '#' or c == '$')

            lookahead = string[i:i + 64]  # NOTE: 64 because of emails

            # Smileys?
            m = SMILEY_RE.match(lookahead)

            if m is not None:
                yield ('smiley', string[i:i + m.end()])
                i += m.end()
                continue

            # Guarding some cases
            if can_be_mention or can_be_hashtag:
                pass

            # Breaking punctuation apart and tokenizing emojis
            elif not c.isalnum() and c != '-':
                j = i + 1

                while j < l:
                    n = string[j]

                    if (
                        n.isspace() or
                        n.isalnum() or
                        is_ascii_junk(n) or
                        n == '#' or
                        n == '@'
                    ):
                        break

                    j += 1

                punct_cluster = string[i:j]

                if punct_cluster in APOSTROPHES:
                    j = i + 1
                else:
                    yield from punct_emoji_iter(punct_cluster)
                    i = j
                    continue

            # Consuming token
            j = i + 1
            token_type = 'word'
            already_consumed = False

            # Hashtags
            if can_be_hashtag:
                if j >= l or not string[j].isalpha():
                    yield ('punct', c)

                    i += 1
                    continue

                token_type = 'hashtag'

                while j < l and string[j].isalnum():
                    j += 1

                already_consumed = True

            # Mentions
            elif can_be_mention:
                if j >= l or not is_valid_twitter_char(string[j]):
                    yield ('punct', c)

                    i += 1
                    continue

                token_type = 'mention'

                while j < l and is_valid_twitter_char(string[j]):
                    j += 1

                already_consumed = True

            # Numerical token
            elif c.isdigit() or c == '-':
                already_consumed = True
                token_type = 'punct'
                last_char = c
                chosen_decimal = None

                while j < l:
                    n = string[j]

                    if n.isspace():
                        break

                    if not n.isdigit():
                        if last_char in DECIMALS:
                            j -= 1
                            break

                        if (
                            last_char != '-' and
                            (
                                n.isalpha() or
                                n in IDENTIFIER_PARTS
                            )
                        ):
                            already_consumed = False
                            token_type = 'word'
                            break

                        elif n not in DECIMALS:
                            break

                        elif chosen_decimal is not None:
                            break

                        else:
                            chosen_decimal = n

                    last_char = n
                    token_type = 'number'
                    j += 1

                if already_consumed and j > i + 1 and string[j - 1] in DECIMALS:
                    j -= 1

            # Alphanumerical token
            can_be_acronym = False

            if not already_consumed:
                could_be_url = starts_as_url(lookahead)
                could_be_email = email_lookahead(lookahead)

                if could_be_url:
                    token_type = 'url'
                elif could_be_email:
                    token_type = 'email'

                while j < l:
                    n = string[j]

                    if n.isspace():
                        break

                    if could_be_url or could_be_email:
                        if n == ',':
                            break

                    elif n in IDENTIFIER_PARTS or n == '·':
                        if j + 1 >= l:
                            break

                        if string[j + 1].isspace():
                            break

                        j += 1
                        continue

                    elif not n.isalnum():

                        # Handling acronyms
                        if n == '.' and string[j - 1].isupper():
                            j += 1
                            can_be_acronym = True
                            continue

                        break

                    j += 1

            before = string[i:j]

            if before[0] in APOSTROPHES and before[1:].lower() not in ENGLISH_ARCHAIC_CONTRACTIONS:
                yield ('punct', before[0])
                i += 1

                if i == j:
                    continue

                before = before[1:]

            if can_be_acronym:
                if all(len(t) <= 1 for t in before.split('.')):
                    yield ('word', before)

                else:
                    for t in POINT_SPLITTER_RE.split(before)[:-1]:
                        yield ('word' if t != '.' else 'punct', t)

                i = j
                continue

            # Handling contractions
            if j < l and string[j] in APOSTROPHES:
                k = j + 1

                while k < l:
                    n = string[k]

                    if not n.isalnum() and n not in SIGILS:
                        break

                    k += 1

                if k > j + 1:
                    after = string[j + 1:k]

                    if after[0].isdigit():
                        yield (token_type, before + string[j])
                        i = j + 1
                        continue

                    elif after in ENGLISH_CONTRACTIONS:
                        yield (token_type, before)
                        yield (token_type, string[j] + after)

                    elif (
                        (before.endswith('n') and after == 't') or
                        after in FRENCH_EXCEPTIONS or
                        '-' in before
                    ):
                        yield (token_type, string[i:k])

                    elif (
                        (
                            after[0] != 'h' and
                            is_consonant(before[-1]) and
                            (is_consonant(after[0]) and after[0] not in SIGILS)
                        ) or
                        before in IRISH
                    ):
                        yield (token_type, before + string[j] + after)

                    else:
                        yield (token_type, before + string[j])
                        i = j + 1
                        continue

                else:
                    yield (token_type, before)
                    yield ('punct', string[j])

                j = k

            # Handling abbreviations
            elif j < l - 1 and string[j] == '.' and before.lower() in ABBREVIATIONS:
                yield (token_type, before + '.')
                j += 1

            # Handling regular word
            else:

                if token_type == 'word':

                    # Handling unions that should be split
                    hyphens = before.count('-')

                    if hyphens == 2 and '-t-' in before:
                        for t in before.split('-'):
                            yield (token_type, t)

                    elif hyphens == 1:
                        a, b = before.split('-')

                        if b in FRENCH_UNION_TARGETS:
                            yield (token_type, a)
                            yield (token_type, b)

                        else:
                            yield (token_type, before)

                    else:
                        yield (token_type, before)

                elif token_type == 'hashtag' and self.split_hashtags:
                    yield from split_hashtag(before)

                else:

                    yield (token_type, before)

            i = j

    def tokenize(self, string):
        if self.decode_html_entities:
            string = unescape(string)

        if self.unidecode:
            string = unidecode_expect_ascii(string)

        if self.reduce_words:
            string = reduce_lenghtening(string)

        if self.__only_defaults:
            yield from self.__tokenize(string)
            return

        for token in self.__tokenize(string):
            token_type, token_value = token
            token_changed = False

            if self.keep is not None and token_type not in self.keep:
                continue

            elif self.drop is not None and token_type in self.drop:
                continue

            elif token_type == 'mention':

                if self.normalize_mentions:
                    token_value = token_value.lower()
                    token_changed = True

                if self.mentions_as_words:
                    token_type = 'word'
                    token_value = token_value[1:]
                    token_changed = True

            elif token_type == 'hashtag':

                if self.normalize_hashtags:
                    token_value = unidecode_expect_ascii(token_value.lower())
                    token_changed = True

                if self.hashtags_as_words:
                    token_type = 'word'
                    token_value = token_value[1:]
                    token_changed = True

            # NOTE: not an `elif` here because the token type can change is hashtags are splitted
            if token_type == 'word':
                if self.lower:
                    token_value = token_value.lower()

                if self.min_word_length is not None and len(token_value) < self.min_word_length:
                    continue

                elif self.max_word_length is not None and len(token_value) > self.max_word_length:
                    continue

                token_changed = True

            if self.stoplist is not None and token_value in self.stoplist:
                continue

            if token_changed:
                token = (token_type, token_value)

            yield token

    def __call__(self, string):
        return self.tokenize(string)



from fog.key.fingerprint import (
    FingerprintingKeyer,
    NgramsFingerprintKeyer
)
from fog.key.levenshtein import (
    levenshtein_1d_keys,
    damerau_levenshtein_1d_keys,
    levenshtein_1d_blocks,
    damerau_levenshtein_1d_blocks,
    levenshtein_2d_blocks,
    damerau_levenshtein_2d_blocks
)
from fog.key.misc import ngram_keys, zig_zag
from fog.key.omission import omission_key
from fog.key.skeleton import skeleton_key

fingerprint = FingerprintingKeyer()
ngrams_fingerprint = NgramsFingerprintKeyer()



# =============================================================================
# Fog Transform CLI Action
# =============================================================================
#
# Logic of the transform CLI action enabling the user to edit cell in batch.
#
import csv
from unidecode import unidecode
from fog.cli.utils import custom_reader
from fog.key import fingerprint
from fog.utils import squeeze

OPERATIONS = {
    'fingerprint': fingerprint,
    'len': len,
    'lower': lambda x: x.lower(),
    'squeeze': squeeze,
    'strip': lambda x: x.strip(),
    'unidecode': unidecode,
    'upper': lambda x: x.upper()
}


def transform_action(namespace):

    # TODO: validate operation_chain
    operation_chain = None

    if not namespace.eval:
        operation_chain = [OPERATIONS[o] for o in namespace.operations.split(',')]

    headers, position, reader = custom_reader(namespace.file, namespace.column)

    if namespace.target_column is not None:
        if namespace.after:
            headers = headers[:position + 1] + [namespace.target_column] + headers[position + 1:]
        else:
            headers.append(namespace.target_column)

    writer = csv.writer(namespace.output)
    writer.writerow(headers)

    for line in reader:
        new_value = line[position]

        if namespace.eval:
            new_value = eval(namespace.operations, None, {
                'unidecode': unidecode,
                'value': new_value
            })
        else:
            for op in operation_chain:
                new_value = op(new_value)

        if namespace.target_column is not None:
            if namespace.after:
                line = line[:position + 1] + [new_value] + line[position + 1:]
            else:
                line.append(new_value)
        else:
            line[position] = new_value

        writer.writerow(line)



# =============================================================================
# Fog Dice Coefficient
# =============================================================================
#
# Functions computing the Sørensen-Dice coefficient.
#
# [Urls]:
# https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient
#
# [References]:
# Dice, Lee R. (1945). "Measures of the Amount of Ecologic Association
# Between Species". Ecology 26 (3): 297–302.
#
from fog.metrics.utils import ACCEPTABLE_TYPES


def dice_coefficient(A, B):
    """
    Function computing the Dice coefficient. That is to say twice the size of
    the intersection of both sets divided by the sum of both their sizes.

    Runs in O(n), n being the size of the smallest set.

    Args:
        A (iterable): First sequence.
        B (iterable): Second sequence.

    Returns:
        float: Dice coefficient between A & B.

    Example:
        from fog.metrics import dice_coefficient

        # Basic
        dice_coefficient('context', 'contact')
        >>> ~0.727

    """
    if A is B:
        return 1.0

    if not isinstance(A, ACCEPTABLE_TYPES):
        A = set(A)

    if not isinstance(B, ACCEPTABLE_TYPES):
        B = set(B)

    # Swapping to iterate over smaller set and minimize lookups
    if len(A) > len(B):
        A, B = B, A

    la = len(A)
    lb = len(B)

    if la == 0:
        return 1.0 if lb == 0 else 0.0

    # Counting intersection
    I = 0

    for v in A:
        if v in B:
            I += 1

    return (2 * I) / (la + lb)



# =============================================================================
# Fog Hashjoin Algorithm
# =============================================================================
#
# TODO: key(s) as iterable of fuctions ?
#
from types import GeneratorType
import warnings


def hashjoin(left, right, key=None, keys=None):
    """
    For each element in left, hasjoin finds similar elements in right.
    Pairs are yieled when two elements share at least one key.

    Returns a list of pair of candidates (left_elt, right_elt).

    Args
        left (iterable): Arbitrary iterable containing data points. Will be
            fully consumed.
        right (iterable): Arbitrary iterable containing data points. Will be
            fully consumed.
        key (function): Functions returning an item's key.
        keys (function): Functions returning an item's keys.

    Yield
        (list): List of pairs found across left and right datasets.
    """
    try:
        if len(left) > len(right):
            A, B = right, left
            invert = True
        else:
            A, B = left, right
            invert = False
    except TypeError:
        A, B = left, right
        invert = False

    if isinstance(B, GeneratorType):
        warnings.warn(
            """
            fog.join.hasjoin: At least one of the input iterables is a
            generator. It will be fully consumed before hasjoin finds all
            similar pairs.
            """)

    # Single key
    if key is not None:
        for a in A:
            key_a = key(a)

            for b in B:
                key_b = key(b)

                if key_a == key_b:

                    if invert:
                        yield (b, a)
                    else:
                        yield (a, b)

    # Multiple keys
    if keys is not None:
        for a in A:
            key_a = keys(a)

            for b in B:
                key_b = keys(b)

                if len(list(set(key_a) & set(key_b))) > 0:

                    if invert:
                        yield (b, a)
                    else:
                        yield (a, b)



# =============================================================================
# Fog LAESA Clustering
# =============================================================================
#
# Implementation of the linear AESA algorithm to perform nearest neighbor
# clustering.
#
# [References]:
# Micó, María Luisa, José Oncina, et Enrique Vidal. « A New Version of the
# Nearest-Neighbour Approximating and Eliminating Search Algorithm (AESA) with
# Linear Preprocessing Time and Memory Requirements ». Pattern Recognition
# Letters 15, no 1 (janvier 1994): 9‑17.
# https://doi.org/10.1016/0167-8655(94)90095-7.
#
# Rico-Juan, Juan Ramón, et Luisa Micó. « Comparison of AESA and LAESA Search
# Algorithms Using String and Tree-Edit-Distances ». Pattern Recognition
# Letters 24, no 9‑10 (juin 2003): 1417‑26.
# https://doi.org/10.1016/S0167-8655(02)00382-3.
#
# [Urls]:
# http://www.xavierdupre.fr/app/mlstatpy/helpsphinx/c_ml/kppv.html
#
# Note that this method is not particularly efficient for the general use
# case. It's often worse than the quadratic approach.
#
import math
import random

from fog.clustering.utils import clusters_from_pairs


def pivot_selection(rng, data, pivots, distance):
    """
    Function in charge of pivot selection.

    """
    b_prime = rng.randint(0, len(data) - 1)
    n = len(data)
    p = set()
    A = [0] * n
    D = []

    # Finding pivots
    for i in range(pivots):
        maximum = 0
        b = b_prime
        D.append([0] * n)
        pivot = data[b]

        for j, item in enumerate(data):
            if j in p:
                continue

            d = distance(pivot, item)
            A[j] += d

            if A[j] > maximum:
                b_prime = j
                maximum = A[j]

        p.add(b_prime)

    p = list(p)

    # Pre-computing distances (not necessary, can be done before or later)
    for i in range(pivots):
        pivot = data[p[i]]

        for j, item in enumerate(data):
            D[i][j] = distance(pivot, item)

    return p, D


# TODO: currently not working
def laesa(data, distance, radius, pivots=None, min_size=2,
          max_size=float('inf'), seed=None, mode='connected_components'):
    """
    Function returning an iterator over found clusters using the LAESA method.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        distance (callable): Distance function to be used.
        radius (number): produced clusters' radius.
        pivots (number, optional): number of pivots to use. Defaults to
            log2(n).
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        seed (number, optional): rng seed. Defaults to None.
        mode (string, optional): 'fuzzy_clusters', 'connected_components'.
            Defaults to 'connected_components'.

    Yield:
        list: A viable cluster.

    """

    # Building rng
    rng = random.Random(seed)

    # We need to consume as a list to be able of random access
    if type(data) is not list:
        data = list(data)

    # Finding the number of pivots
    if pivots is None:
        pivots = math.ceil(math.log2(len(data)))

    # Finding good pivots
    p, D = pivot_selection(rng, data, pivots, distance)
    n = len(data)

    def clustering():
        DD = 0
        for i, x in enumerate(data):

            # Distance to the nearest pivot
            d_star = float('inf')

            for j in range(pivots):
                d = distance(data[p[j]], x)

                if d < d_star:
                    d_star = d

            for k in range(i + 1, n):

                compute = True

                for j in range(pivots):
                    delta = abs(D[j][i] - D[j][k])

                    if delta > d_star:
                        compute = False
                        break

                if compute:
                    d = distance(x, data[k])
                    DD += 1

                    if d <= radius:
                        yield (x, data[k])

        print('DISTANCES: %i' % DD)

    yield from clusters_from_pairs(
        clustering(),
        min_size=min_size,
        max_size=max_size,
        mode=mode
    )



# =============================================================================
# Fog Miscellaneaous Keys
# =============================================================================
#
# Miscellaneous set of key functions used for blocking and/or sorted
# neighborhood approaches.
#
# [References]:
# Christen, Peter. « A Survey of Indexing Techniques for Scalable Record
# Linkage and Deduplication ». IEEE Transactions on Knowledge and Data
# Engineering 24, no 9 (septembre 2012): 1537‑55.
# https://doi.org/10.1109/TKDE.2011.127.
#
# Knuth, D. 1997. The Art of Computer Programming. Sorting and Searching,
# 2d ed., vol. 3. Addison-Wesley, Upper Saddle River, NJ. p. 394
#
import math
from itertools import chain, combinations
from fog.tokenizers import ngrams

# Following Knuth's intuitions, sorting keys twice, once normally and once
# over the reversed strings, yields very good results when applying sorted
# neighborhood method over edit distances.
zig_zag = (None, lambda x: x[::-1])


# TODO: way to choose l instead in negative by number max of match
def ngram_keys(n, string, threshold=0.8):
    """
    Function returning n-gram keys for the given string. Useful for blocking.
    Note that this method is only suitable with small strings (a single name,
    for instance), else the number of combinations will explode and generate
    way too many keys.

    Args:
        n (number): Size of the grams.
        string (str): Target string.
        threshold (number, optional): Threshold controlling the number
            of grams that can be missing from two compared strings so they
            collide on at least one key. Defaults to 0.8.

    Yields:
        str: A single key.

    """
    k = len(string) - (n - 1)
    l = max(1, math.floor(k * threshold))

    grams = list(ngrams(n, string))

    yield ''.join(grams)

    s = k
    iterator = iter([])

    while s > l:
        s -= 1

        iterator = chain(iterator, combinations(grams, s))

    for g in iterator:
        yield ''.join(g)



# =============================================================================
# Fog Levensthein Keys
# =============================================================================
#
# Functions related to solving the Levenshtein & Damerau-Levenshtein <= 1/2
# problem efficiently by providing collision keys & blocks.
#
# [References]:
# Boytsov, Leonid. « Indexing Methods for Approximate Dictionary Searching:
# Comparative Analysis ». Journal of Experimental Algorithmics 16 (1 mai 2011).
# https://doi.org/10.1145/1963190.1963191.
#
# Doster. « Contextual Postprocessing System for Cooperation with a
# Multiple-Choice Character-Recognition System ». IEEE Transactions on
# Computers C-26, no 11 (novembre 1977): 1090‑1101.
# https://doi.org/10.1109/TC.1977.1674755.
#
# Knuth, D. 1997. The Art of Computer Programming. Sorting and Searching,
# 2d ed., vol. 3. Addison-Wesley, Upper Saddle River, NJ. p. 394
#
# Sutinen, Erkki, et Jorma Tarhio. « Filtration with q-samples in approximate
# string matching ». In Combinatorial Pattern Matching, 1075:50‑63. Berlin,
# Heidelberg: Springer Berlin Heidelberg, 1996.
# https://doi.org/10.1007/3-540-61258-0_4.
#
# Bast, Hannah, et Marjan Celikik. « Efficient Fuzzy Search in Large Text
# Collections ». ACM Transactions on Information Systems 31, no 2
# (1 mai 2013): 1‑59.
# https://doi.org/10.1145/2457465.2457470.
#
from functools import partial


# TODO: also try Mor-Fraenkel index for k = 1 using only deletions
# TODO: k-truncated from Efficient Fuzzy Search in Large Text Collections.pdf
# TODO: rename wildcard_neighborhood and rename package
def levenshtein_1d_keys(string, transpositions=False, flag='\x00'):
    """
    Function returning an iterator over Levenshtein 1D keys, being the series
    of keys colliding with other strings being at a Levenshtein distance of
    1 with the given string.

    This works by generating a set of collision keys such as two strings may
    only share a key if they have a Levenshtein distance <= 1.

    Note that the number of keys is proportional to the size of the string:
        - 1 key + n substition keys + (n + 2) addition keys, i.e. (2n + 3) keys
          for Levenshtein distance.
        - n - 1 keys for transpositions, i.e. (3n + 2) total keys
          for Damerau-Levenshtein distance.

    In the literature, this technique seems to be called generating the
    "wildcard neighborhood" of a string. It's quite efficient compared to
    traditional neighborhood methods because the number of keys is not
    dependent on the size of the alphabet since we use a wildcard to represent
    the substitutions/additions.

    Those keys are useful to perform O(n) clustering with Levenshtein
    distance <= 1 but can become very costly with long strings since the number
    of produced keys is a factor of your string's length.

    Args:
        string (string): Target string
        transpositions (bool, optional): Whether to support transpositions
            like with the Damerau-Levenshtein distance. Defaults to False.

    Yields:
        string: A single Levenshtein 1D key.

    """

    n = len(string)

    yield string

    for i in range(n):

        # Substitution
        yield string[:i] + flag + string[i + 1:]

        # Transpositions
        if i > 0 and transpositions:
            if string[i - 1] == string[i]:
                continue

            yield string[:i - 1] + string[i] + string[i - 1] + string[i + 1:]

        # Additions
        # Note: I use additions here and not deletions because indexing
        # on deletions can generate false positives such as:
        # "abcd", "abde" -> key "abd".
        # Note: for k = 1, an addition for A->B is compulsorily a deletion
        # for B->A. So, in fact, substitution key and transposition keys will
        # match when required.

        # When some letters are doubled, we can skip some keys
        if i > 0 and string[i - 1] == string[i]:
            continue

        yield string[:i] + flag + string[i:]

    # Last addition
    yield string + flag


damerau_levenshtein_1d_keys = partial(levenshtein_1d_keys, transpositions=True)


# TODO: half-grams, easily parallelizable method
# TODO: include positional info + length info (get filtering refs in paper)
# TODO: no rolling to avoid generating too much keys
# TODO: test the deletion of middle char (more collisions, less keys)
# TODO: if even and transposing, if doubled letter, don't need to generate tr keys
def levenshtein_1d_blocks(string, transpositions=False, flag='\x00'):
    """
    Function returning the minimal set of longest Levenshtein distance <= 1
    blocking keys of target string. Under the hood, this splits the given
    string into an average of 3 blocks (2 when string length is even, 4 when
    odd). When supporting transpositions, the function will always return
    4 blocks to handle the case when middle letters are transposed.

    Note that this method therefore yields a constant number of keys, as
    opposed to ngrams which yield a number of keys proportional to the size
    of the given strings.

    The intuition of this technique can be found in Knuth. Similar concepts can
    be found in Doster. This is also reminiscent of q-samples.

    This method works quite well because it's independent of the string's
    length and therefore implicitly incorporates the string's length into
    the generated keys.

    Indeed, this method minimizes the number of produced keys (the number
    is constant, contrary to n-grams, for instance) and minimizes the
    probability of two strings colliding.

    What's more, this method is exact and won't generate false negatives like
    n-grams would.

    Args:
        string (str): Target string.
        transpositions (bool, optional): Whether to support transpositions
            like with the Damerau-Levenshtein distance. Defaults to False.

    Returns:
        tuple: Resulting blocks.

    """
    n = len(string)

    if n == 1:
        return (flag + string, string + flag, flag)

    h = n // 2

    # String has even length, we just split in half
    if n % 2 == 0:

        first_half = flag + string[:h]
        second_half = string[h:] + flag

        # For transpositions, we swap once the middle characters
        if transpositions:

            first_transposed_half = flag + string[:h - 1] + second_half[0]
            second_transposed_half = first_half[-1] + string[h + 1:] + flag

            return (first_half, second_half, first_transposed_half, second_transposed_half)
        else:
            return (first_half, second_half)

    # String has odd length, we split twice
    h1 = h + 1

    first_half = flag + string[:h]
    second_half = string[h:] + flag
    first_half1 = flag + string[:h1]
    second_half1 = string[h1:] + flag

    return (first_half, second_half, first_half1, second_half1)


damerau_levenshtein_1d_blocks = partial(levenshtein_1d_blocks, transpositions=True)


# TODO: generalize to 3 and make a dynamic version
def levenshtein_2d_blocks(string, transpositions=False,
                          flag1='\x00', flag2='\x01', flag3='\x02'):
    """
    Function returning the minimal set of longest Levenshtein distance <= 2
    blocking keys of target string. Note that this method is basically a
    generalization of the `levensthein_1d_blocks` scheme.

    Args:
        string (str): Target string.
        transpositions (bool, optional): Whether to support transpositions
            like with the Damerau-Levenshtein distance. Defaults to False.

    Returns:
        tuple: Resulting blocks.

    """

    n = len(string)
    q = n % 3

    if q == 0:
        h = n // 3

        basic = (
            flag1 + string[:h],
            flag2 + string[h:h * 2],
            flag3 + string[h * 2:]
        )

        if transpositions:
            h1 = h - 1

            return basic + (
                flag1 + string[:h1] + string[h],
                flag2 + string[h1] + string[h + 1:h * 2],
                flag3 + string[h * 2 - 1] + string[h * 2 + 1:]
            )

        return basic

    elif q == 2:
        h1 = n // 3
        h = h1 + 1

        return (
            flag1 + string[:h],
            flag2 + string[h:h * 2],
            flag3 + string[h * 2:],

            flag1 + string[:h1],
            flag2 + string[h1:h + h1],
            flag3 + string[h + h1:],

            flag2 + string[h:h + h1]
        )

    elif q == 1:
        h1 = n // 3
        h = h1 + 1

        return (
            flag1 + string[:h],
            flag2 + string[h:h + h1],
            flag3 + string[h + h1:],

            flag1 + string[:h1],
            flag2 + string[h1:h1 + h],

            flag2 + string[h1:h1 * 2],
            flag3 + string[h1 * 2:]
        )


damerau_levenshtein_2d_blocks = partial(levenshtein_2d_blocks, transpositions=True)



# =============================================================================
# Fog Utility Functions
# =============================================================================
#
# Miscellaneous functions used throughout the various module.
#
import re
from statistics import StatisticsError

SQUEEZE_RE = re.compile(r'(.)\1+')
SQUEEZE_ROMAN_RE = re.compile(r'([^IXCMixcm])\1+')


def squeeze(string, keep_roman_numerals=False):
    """
    Simple helper dropping consecutive duplicates in a string.

    Args:
        string (string): Target string.
        keep_roman_numerals (boolean, optional): Whether not to squeeze roman
            numerals. Defaults to `False`.

    Returns:
        string: The string, without consecutive duplicates.

    """
    return (SQUEEZE_ROMAN_RE if keep_roman_numerals else SQUEEZE_RE).sub(r'\1', string)


class OnlineMean(object):
    """
    Just a helper class to compute the mean of a stream of values online and
    only using constant memory.

    Note that it might be subject to float precision issues.
    """

    __slots__ = ('_sum', '_n')

    def __init__(self):
        self._sum = 0
        self._n = 0

    def add(self, x):
        self._n += 1
        self._sum += x

    def subtract(self, x):
        assert self._n != 0
        self._n -= 1
        self._sum -= x

    def __add__(self, other):
        new = self.__class__()
        new.__iadd__(self)
        new.__iadd__(other)

        return new

    def __iadd__(self, other):
        if not isinstance(other, OnlineMean):
            raise TypeError

        self._n += other._n
        self._sum += other._sum

        return self

    def __len__(self):
        return self._n

    def peek(self):
        if self._n == 0:
            raise StatisticsError

        return self._sum / self._n

    def __float__(self):
        return self.peek()



# =============================================================================
# Fog Split CLI Action
# =============================================================================
#
# Logic of the split CLI action that can split multi-valued column cells
# into different lines.
#
import csv

from fog.cli.utils import custom_reader


def split_action(namespace):
    headers, position, reader = custom_reader(namespace.file, namespace.column)

    if namespace.target_column is not None:
        headers.append(namespace.target_column)

    if namespace.rename_column is not None:
        headers[position] = namespace.rename_column

    writer = csv.writer(namespace.output)
    writer.writerow(headers)

    for line in reader:
        values = line[position]

        if len(values) == 0:
            writer.writerow(line)
            continue

        for value in values.split(namespace.separator):

            if namespace.target_column:
                line.append(value)
            else:
                line[position] = value

            writer.writerow(line)



# =============================================================================
# Fog Pairwise Clustering
# =============================================================================
#
# Clustering algorithms computing every pairwise distance/similarity to build
# suitable clusters.
#
import dill
from collections import defaultdict
from multiprocessing import Pool
from phylactery import UnionFind

from fog.clustering.utils import (
    make_similarity_function,
    upper_triangular_matrix_chunk_iter
)


def pairwise_leader(data, similarity=None, distance=None, radius=None,
                    min_size=2, max_size=float('inf')):
    """
    Function returning an iterator over found clusters using the leader
    algorithm.

    It works by assuming the similarity relation is transitive, meaning that
    (A =~ B, B =~ C) leads to A =~ C, which is often quite correct.

    Under the hood, each point will start in its own cluster and each time
    two points are deemed similar, they are merged into the same cluster. Then,
    the first item of the cluster, its leader if you will, is the only one
    that will be used in subsequent similarity comparisons.

    This means three things:
        1) The radius of the produced clusters is sound.
        2) The items' order can have an arbitrary influence on the produced
           clusters.
        3) One item CANNOT belong to more than one cluster.

    This algorithm runs in O(n * (c - 1) / 2), n being the number of items and c
    being the number of clusters, i.e. O(n^2) in practice since for record
    linkage most items will be alone in their clusters.

    Note that this algorithm can work by storing only the current cluster in
    memory.

    Example:
        The following chain:
            ('abc', 'bcd', 'cde', 'def', 'efg', 'ghi')
        will produce the following levenshtein radius=2 clusters:
            ('abc', 'bcd')
            ('cde', 'def')
            ('efg', 'ghi')

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        similarity (callable): If radius is specified, a function returning
            the similarity between two points. Else, a function returning
            whether two points should be deemed similar. Alternatively, one can
            specify `distance` instead.
        distance (callable): If radius is specified, a function returning
            the distance between two points. Else, a function returning
            whether two point should not be deemed similar. Alternatively, one
            can specify `similarity` instead.
        radius (number, optional): produced clusters' radius.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.

    Yields:
        list: A viable cluster.

    """

    # Formatting similarity
    similarity = make_similarity_function(similarity=similarity, distance=distance, radius=radius)

    # We need to consume as a list to be able of random access
    if type(data) is not list:
        data = list(data)

    n = len(data)
    visited = set()

    for i in range(n):
        if i in visited:
            continue

        A = data[i]

        cluster = None

        for j in range(i + 1, n):
            if j in visited:
                continue

            B = data[j]

            if similarity(A, B):

                if cluster is None:
                    cluster = [A, B]
                else:
                    cluster.append(B)

                if len(cluster) > max_size:
                    for c in range(1, len(cluster)):
                        visited.delete(c)
                    break

                visited.add(j)

        if cluster is not None and len(cluster) >= min_size:
            yield cluster


def pairwise_worker(payload):
    """
    Worker function used to compute pairwise computations over chunks of
    an upper triangular matrix in parallel.

    """
    similarity, I, J, offset_i, offset_j = payload

    similarity = dill.loads(similarity)
    pairs = []

    diagonal_chunk = offset_i == offset_j

    if diagonal_chunk:
        J = I

    for i in range(len(I)):
        A = I[i]

        for j in range(0 if not diagonal_chunk else i + 1, len(J)):
            B = J[j]

            if similarity(A, B):
                pairs.append((offset_i + i, offset_j + j))

    return pairs


def pairwise_fuzzy_clusters(data, similarity=None, distance=None, radius=None,
                            min_size=2, max_size=float('inf'), key=None,
                            processes=1, chunk_size=100):
    """
    Function returning an iterator over found clusters using an algorithm
    yielding fuzzy clusters.

    This algorithm is a compromise between the "leader" method and the
    "connected components" one in that it won't produce clusters with a radius
    larger that what's expected and will not attempt to cut clusters too
    arbitrarily. To achieve that, this algorithm will yield fuzzy clusters.

    This means three things:
        1) The radius of the produced clusters is sound.
        2) The items' order can have an arbitrary influence on the produced
           clusters.
        3) One item CAN belong to more than one cluster

    This algorithm runs in O(n * (n - 1) / 2), i.e. O(n^2).

    Note that this algorithm can be parallelized and then run in
    O (n * (n - 1) / 2 / p), p being the number of processes.

    TODO: option to sort by degree

    Example:
        The following chain:
            ('abc', 'bcd', 'cde', 'def', 'efg', 'ghi')
        will produce the following levenshtein radius=2 clusters:
            ('abc', 'bcd')
            ('cde', 'bcd', 'def')
            ('efg', 'def', 'fgh')
            ('ghi', 'fgh')

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        similarity (callable): If radius is specified, a function returning
            the similarity between two points. Else, a function returning
            whether two points should be deemed similar. Alternatively, one can
            specify `distance` instead.
        distance (callable): If radius is specified, a function returning
            the distance between two points. Else, a function returning
            whether two point should not be deemed similar. Alternatively, one
            can specify `similarity` instead.
        radius (number, optional): produced clusters' radius.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        key (callable, optional): function returning an item's key.
            Defaults to None.
        processes (number, optional): number of processes to use. Defaults to 1.
        chunk_size (number, optional): size of matrix chunks to send to
            subprocesses. Defaults to 100.

    Yields:
        list: A viable cluster.

    """

    # Formatting similarity
    similarity = make_similarity_function(similarity=similarity, distance=distance, radius=radius)

    # We need to consume as a list to be able of random access
    if type(data) is not list:
        data = list(data)

    if key is not None:
        keys = list(key(item) for item in data)
    else:
        keys = data

    n = len(data)
    graph = defaultdict(list)

    # Computing similarities
    if processes == 1:
        for i in range(n):
            A = keys[i]

            for j in range(i + 1, n):
                B = keys[j]

                if similarity(A, B):
                    graph[i].append(j)
                    graph[j].append(i)
    else:

        pickled_similarity = dill.dumps(similarity)

        # Iterator
        pool_iter = (
            (pickled_similarity, ) + chunk
            for chunk
            in upper_triangular_matrix_chunk_iter(keys, chunk_size)
        )

        # Pool
        with Pool(processes=processes) as pool:
            for matches in pool.imap(pairwise_worker, pool_iter):
                for i, j in matches:
                    graph[i].append(j)
                    graph[j].append(i)

    # Building clusters
    visited = set()
    for i, neighbors in graph.items():
        if i in visited:
            continue

        if len(neighbors) + 1 < min_size:
            continue
        if len(neighbors) + 1 > max_size:
            continue

        visited.update(neighbors)

        cluster = [data[i]] + [data[j] for j in neighbors]
        yield cluster


def pairwise_connected_components(data, similarity=None, distance=None, radius=None,
                                  min_size=2, max_size=float('inf'), key=None,
                                  processes=1, chunk_size=100):
    """
    Function returning an iterator over found clusters by computing a
    similarity graph of given data and extracting its connected components.

    It works by assuming the similarity relation is transitive, meaning that
    (A =~ B, B =~ C) leads to A =~ C, which is often quite correct.

    This means two things:
        1) The produced clusters may have a larger radius that what's
           intended.
        2) One item cannot belong to more than one cluster.

    This algorithm runs in O(n * (c - 1) / 2), n being the number of items and
    c being the number of connected components, i.e. O(n^2) in practice since
    for record linkage most items will be alone in their clusters.

    Example:
        The following chain:
            ('abc', 'bcd', 'cde', 'def', 'efg', 'ghi')
        will produce the following levenshtein radius=2 cluster:
            ('abc', 'bcd', 'cde', 'def', 'efg', 'ghi')

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        similarity (callable): If radius is specified, a function returning
            the similarity between two points. Else, a function returning
            whether two points should be deemed similar. Alternatively, one can
            specify `distance` instead.
        distance (callable): If radius is specified, a function returning
            the distance between two points. Else, a function returning
            whether two point should not be deemed similar. Alternatively, one
            can specify `similarity` instead.
        radius (number, optional): produced clusters' radius.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        key (callable, optional): function returning an item's key.
            Defaults to None.
        processes (number, optional): number of processes to use. Defaults to 1.
        chunk_size (number, optional): size of matrix chunks to send to
            subprocesses. Defaults to 100.

    Yields:
        list: A viable cluster.

    """

    # Formatting similarity
    similarity = make_similarity_function(similarity=similarity, distance=distance, radius=radius)

    # We need to consume as a list to be able of random access
    if type(data) is not list:
        data = list(data)

    if key is not None:
        keys = list(key(item) for item in data)
    else:
        keys = data

    n = len(data)
    sets = UnionFind(n)

    # Computing pairwise distances
    if processes == 1:
        for i in range(n):
            A = keys[i]

            # NOTE: if one adds a cardinality test hear, we get the leader algo

            for j in range(i + 1, n):
                B = keys[j]

                if sets.connected(i, j):
                    continue

                if similarity(A, B):
                    sets.union(i, j)
    else:
        pickled_similarity = dill.dumps(similarity)

        # Iterator
        pool_iter = (
            (pickled_similarity, ) + chunk
            for chunk
            in upper_triangular_matrix_chunk_iter(keys, chunk_size)
        )

        # Pool
        with Pool(processes=processes) as pool:
            for matches in pool.imap_unordered(pairwise_worker, pool_iter):
                for i, j in matches:
                    sets.union(i, j)

    # TODO: Should really be using the sparse version
    for component in sets.components(min_size=min_size, max_size=max_size):
        yield [data[i] for i in component]


def pairwise(data, mode='connected_components', **kwargs):
    """
    Function returning an iterator over found clusters by computing pairwise
    similarity or distance.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        mode (str): 'leader' or 'fuzzy_clusters' or 'connected_components'.
            Defaults to 'connected_components'.
        similarity (callable): If radius is specified, a function returning
            the similarity between two points. Else, a function returning
            whether two points should be deemed similar. Alternatively, one can
            specify `distance` instead.
        distance (callable): If radius is specified, a function returning
            the distance between two points. Else, a function returning
            whether two point should not be deemed similar. Alternatively, one
            can specify `similarity` instead.
        radius (number, optional): produced clusters' radius.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.

    Yields:
        list: A viable cluster.

    """
    if mode not in ['leader', 'fuzzy_clusters', 'connected_components']:
        raise TypeError('fog.clustering.pairwise: wrong mode "%s"' % mode)

    if mode == 'leader':
        return pairwise_leader(data, **kwargs)

    if mode == 'fuzzy_clusters':
        return pairwise_fuzzy_clusters(data, **kwargs)

    return pairwise_connected_components(data, **kwargs)



# =============================================================================
# Fog PPJoin Clustering
# =============================================================================
#
# Implementation of the PPJoin algorithm.
#
# [References]:
# Xiao, Chuan, Wei Wang, Xuemin Lin, Jeffrey Xu Yu, et Guoren Wang.
# « Efficient Similarity Joins for Near-Duplicate Detection ». ACM Transactions
# on Database Systems 36, no 3 (1 août 2011): 1‑41.
# https://doi.org/10.1145/2000824.2000825.
# https://web.archive.org/web/20170624090246/http://www.cse.unsw.edu.au/~weiw/project/simjoin.html
#
# https://github.com/TsinghuaDatabaseGroup/Similarity-Search-and-Join
# http://www.vldb.org/pvldb/vol9/p636-mann.pdf
#
import math
from collections import defaultdict, Counter
from bisect import bisect_left
from ebbe import sorted_uniq

from fog.lsh.utils import crc32

EPSILON = 1e-6
PRUNE_FLAG = -1
MAX_DEPTH = 2

TOKEN_ORDERINGS = ('freq', 'crc32')


class MetricHelper(object):
    __slots__ = ['threshold']

    def __init__(self, threshold):
        self.threshold = threshold


class JaccardHelper(MetricHelper):
    def index_length(self, l):
        return int((1 - 2 * self.threshold / (1 + self.threshold)) * l + 1 + EPSILON)

    def probe_length(self, l):
        return int((1 - self.threshold) * l + 1 + EPSILON)

    def require_overlap(self, l1, l2):
        return math.ceil(self.threshold / (1 + self.threshold) * (l1 + l2) - EPSILON)

    def compute_similarity(self, l1, l2, overlap):
        return overlap / (l1 + l2 - overlap) + EPSILON

    def max_possible_length(self, l):
        return int(l / self.threshold + EPSILON)

    def min_possible_length(self, l):
        return math.ceil(l * self.threshold - EPSILON)


class DiceHelper(MetricHelper):
    def index_length(self, l):
        return int((1 - self.threshold) * l + 1 + EPSILON)

    def probe_length(self, l):
        return int((1 - self.threshold / (2 - self.threshold)) * l + 1 + EPSILON)

    def require_overlap(self, l1, l2):
        return math.ceil((l1 + l2) * self.threshold / 2.0 - EPSILON)

    def compute_similarity(self, l1, l2, overlap):
        return 2 * overlap / (l1 + l2) + EPSILON

    def max_possible_length(self, l):
        return int(l / self.threshold * (2 - self.threshold) + EPSILON)

    def min_possible_length(self, l):
        return math.ceil(l * self.threshold / (2 - self.threshold) - EPSILON)


class BinaryCosineHelper(MetricHelper):
    def index_length(self, l):
        return int((1 - self.threshold) * l + 1 + EPSILON)

    def probe_length(self, l):
        return int((1 - self.threshold * self.threshold) * l + 1 + EPSILON)

    def require_overlap(self, l1, l2):
        return math.ceil(self.threshold * math.sqrt(l1 * l2) - EPSILON)

    def compute_similarity(self, l1, l2, overlap):
        return overlap / math.sqrt(l1 * l2) + EPSILON

    def max_possible_length(self, l):
        return int(l / self.threshold / self.threshold + EPSILON)

    def min_possible_length(self, l):
        return math.ceil(l * self.threshold * self.threshold - EPSILON)


class InvertedIndexItem(object):
    __slots__ = ('pos', 'ids')

    def __init__(self):
        self.pos = 0
        self.ids = []

    def __iter__(self):
        yield self.pos
        yield self.ids


def compute_overlap(x, y, require_overlap):
    lx = len(x)
    ly = len(y)

    posx = 0
    posy = 0
    current_overlap = 0

    while posx < lx and posy < ly:
        if (
            lx - posx + current_overlap < require_overlap
            or ly - posy + current_overlap < require_overlap
        ):
            return -1

        if x[posx] == y[posy]:
            current_overlap += 1
            posx += 1
            posy += 1
        elif x[posx] < y[posy]:
            posx += 1
        else:
            posy += 1

    return current_overlap


def suffix_filter(x, y, x_start, x_end, y_start, y_end, hd, depth=0):
    x_len = x_end - x_start
    y_len = y_end - y_start

    if x_end <= x_start or y_end <= y_start:
        return abs(x_len - y_len)

    left = 0
    right = 0
    offset = 0

    hd_left = 0
    hd_right = 0
    hd_left_bound = 0
    hd_right_bound = 0

    mid = x_start + x_len // 2
    token = x[mid]

    if x_len >= y_len:
        delta = x_len - y_len
        offset = (hd - delta) // 2 + delta
        left = y_start + x_len // 2 - offset
        offset = (hd - delta) // 2
        right = y_start + x_len // 2 + offset
    else:
        delta = y_len - x_len
        offset = (hd - delta) // 2
        left = y_start + x_len // 2 - offset
        offset = (hd - delta) // 2 + delta
        right = y_start + x_len // 2 + offset

    if (
        (left >= y_start and y[left] > token)
        or (right < y_end and y[right] < token)
    ):
        return hd + 1

    search_left = left if left >= y_start else y_start
    search_right = right + 1 if right + 1 < y_end else y_end

    pos = bisect_left(y, token, search_left, search_right)

    if pos < y_end and y[pos] == token:
        hd_left = hd_left_bound = abs((mid - x_start) - (pos - y_start))
        hd_right = hd_right_bound = abs((x_end - mid - 1) - (y_end - pos - 1))

        if hd_left_bound + hd_right_bound > hd:
            return hd_left_bound + hd_right_bound

        if depth < MAX_DEPTH:
            hd_left = suffix_filter(x, y, x_start, mid, y_start, pos, hd - hd_right_bound, depth + 1)

            if hd_left + hd_right_bound > hd:
                return hd_left + hd_right_bound

            hd_right = suffix_filter(x, y, mid + 1, x_end, pos + 1, y_end, hd - hd_left, depth + 1)

        return hd_left + hd_right
    else:
        hd_left = hd_left_bound = abs((mid - x_start) - (pos - y_start))
        hd_right = hd_right_bound = abs((x_end - mid - 1) - (y_end - pos))

        if hd_left_bound + hd_right_bound + 1 > hd:
            return hd_left_bound + hd_right_bound + 1

        if depth < MAX_DEPTH:
            hd_left = suffix_filter(x, y, x_start, mid, y_start, pos, hd - hd_right_bound - 1, depth + 1)

            if hd_left + hd_right_bound + 1 > hd:
                return hd_left + hd_right_bound + 1

            hd_right = suffix_filter(x, y, mid + 1, x_end, pos, y_end, hd - hd_left - 1, depth + 1)

        return hd_left + hd_right + 1

    return 0


def preprocess(records, tokenizer=None, token_ordering=None):
    freqs = Counter()
    tokenized_records = []

    for record in records:
        if tokenizer is not None:
            record = tokenizer(record)

            if not isinstance(record, list):
                record = list(record)

            if token_ordering == 'freq':
                for token in record:
                    freqs[token] += 1

        record = sorted_uniq(record)

        if token_ordering == 'crc32':
            record = sorted(crc32(token) for token in record)

        tokenized_records.append(record)

    if token_ordering == 'freq':
        labels = {token: i for i, (token, _) in enumerate(sorted(freqs.items(), key=lambda x: (x[1], x[0])))}

        tokenized_records = [
            sorted(labels[token] for token in record)
            for record in tokenized_records
        ]

    argsort = sorted(range(len(tokenized_records)), key=lambda i: len(tokenized_records[i]))

    return tokenized_records, argsort


def ppjoin(records, threshold, metric='jaccard', tokenizer=None, all_pairs=False,
           plus=False, token_ordering='freq'):
    """
    Function returning an iterator over similar pairs of records found using
    the All-Pairs, PPJoin or PPJoin+ algorithm.

    Args:
        records (iterable): The records to work on.
        threshold (float): The similarity threshold under which pairs of
            records won't be deemed similar enough.
        metric (str, optional): The similarity metric to use. Can be `jaccard`,
            `dice` or `binary_cosine`. Defaults to `jaccard`.
        tokenizer (callable, optional): An optional tokenizer function processing
            the records, such as ngrams etc. Defaults to `None`.
        all_pairs (bool, optional): Whether to avoid leveraging PPJoin's
            filtering strategies and only apply the simpler All-Pairs
            algorithm instead. Defaults to `False`.
        plus (bool, optional): Whether to use PPJoin+'s suffix filtering.
            Defaults to `False`.
        token_ordering (str, optional): Which kind of token global ordering
            to use when sorting tokens for prefix filtering. Can be `None`,
            `freq` or `crc32`. `None` mean sorting will be done alphabetically.
            `freq` will sort tokens by increasing corpus frequency to
            minimize collisions. Finally `crc32` will map tokens to their crc32
            hash, meaning the order will be random. Defaults to `freq`.

    Yields:
        tuple: A similar pair.

    """

    if tokenizer is not None and not callable(tokenizer):
        raise TypeError('fog.clustering.ppjoin: tokenizer is not callable')

    if token_ordering is not None and token_ordering not in TOKEN_ORDERINGS:
        raise TypeError('fog.clustering.ppjoin: unknown token ordering "%s"' % token_ordering)

    # Instantiating metric helper
    if metric == 'jaccard':
        helper = JaccardHelper(threshold)
    elif metric == 'dice':
        helper = DiceHelper(threshold)
    elif metric == 'binary_cosine':
        helper = BinaryCosineHelper(threshold)
    else:
        raise TypeError('fog.clustering.ppjoin: unsupported metric "%s"' % metric)

    # First we need to order records by length and make them indexable
    if not isinstance(records, list):
        records = list(records)

    tokenized_records, argsort = preprocess(records, tokenizer, token_ordering)

    # State
    inverted_index = defaultdict(InvertedIndexItem)

    # Performing the join
    for k in argsort:
        record = tokenized_records[k]
        record_length = len(record)

        min_length = helper.min_possible_length(record_length)
        probe_length = helper.probe_length(record_length)
        index_length = helper.index_length(record_length)

        # NOTE: maybe we could try avoiding storing this in memory?
        require_overlaps = [0] * (record_length + 1)

        for l in range(min_length, record_length + 1):
            require_overlaps[l] = helper.require_overlap(record_length, l)

        occurances = {}

        for t in range(0, probe_length):
            token = record[t]
            index_item = inverted_index[token]
            ids = index_item.ids

            while index_item.pos < len(ids) and len(tokenized_records[ids[index_item.pos][0]]) < min_length:
                index_item.pos += 1

            for p in range(index_item.pos, len(ids)):
                candidate_id = ids[p][0]

                if all_pairs:
                    if candidate_id not in occurances:
                        occurances[candidate_id] = 1
                    else:
                        occurances[candidate_id] += 1

                    continue

                candidate_pos = ids[p][1]
                candidate_length = len(tokenized_records[candidate_id])

                require_overlap = require_overlaps[candidate_length]

                value = occurances.get(candidate_id)

                # NOTE: this should be here in principle, but it does not change much?
                if value == PRUNE_FLAG:
                    continue

                # Suffix filtering
                if plus:
                    hamming_distance = (
                        candidate_length +
                        record_length -
                        2 * require_overlap +
                        (0 if value is None else 2 * value) -
                        (candidate_pos + t)
                    )

                    # NOTE: I am not completely sure I am not missing something here...
                    if hamming_distance >= 0:
                        sf = suffix_filter(
                            tokenized_records[candidate_id],
                            record,
                            candidate_pos,
                            candidate_length,
                            t,
                            record_length,
                            hamming_distance
                        )

                        if sf > hamming_distance:
                            if value is not None:
                                occurances[candidate_id] = PRUNE_FLAG

                            continue

                if value is None:
                    if (
                        record_length - threshold < require_overlap
                        or (candidate_length - candidate_pos) < require_overlap
                    ):
                        continue

                    occurances[candidate_id] = 1
                else:
                    if (
                        value + (record_length - threshold) < require_overlap
                        or value + (candidate_length - candidate_pos) < require_overlap
                    ):
                        occurances[candidate_id] = PRUNE_FLAG
                    else:
                        occurances[candidate_id] += 1

            if threshold < index_length:
                ids.append((k, t))

        for candidate, count in occurances.items():
            if count == PRUNE_FLAG:
                continue

            candidate_record = tokenized_records[candidate]
            candidate_size = len(candidate_record)
            require_overlap = require_overlaps[candidate_size]
            index_length = helper.index_length(candidate_size)

            if candidate_record[index_length - 1] < record[probe_length - 1]:
                if count + candidate_size - index_length < require_overlap:
                    continue
            else:
                if count + record_length - probe_length < require_overlap:
                    continue

            real_overlap = compute_overlap(record, candidate_record, require_overlap)

            if real_overlap == -1:
                continue

            similarity = helper.compute_similarity(record_length, candidate_size, real_overlap)

            if similarity >= threshold:
                yield records[k], records[candidate]


def all_pairs(records, threshold, metric='jaccard', tokenizer=None,
              token_ordering='freq'):
    return ppjoin(
        records,
        threshold,
        metric=metric,
        tokenizer=tokenizer,
        token_ordering=token_ordering,
        all_pairs=True
    )


def ppjoin_plus(records, threshold, metric='jaccard', tokenizer=None,
                token_ordering='freq'):
    return ppjoin(
        records,
        threshold,
        metric=metric,
        tokenizer=tokenizer,
        token_ordering=token_ordering,
        plus=True
    )



# =============================================================================
# Fog Vantage Point Tree Clustering
# =============================================================================
#
# Clustering algorithm leveraging a Vantage Point Tree to avoid needing to
# compute pairwise distances. In practice, it means that the number of
# distance computations needed per point will be sublinear. Therefore,
# parallelization of the brute-force algorithm often yields better results.
#
# [Reference]:
# Data Structures and Algorithms for Nearest Neighbor Search in General
# Metric Space. Peter N. Yianilos.
#
from phylactery import VPTree


def vp_tree(data, distance, radius, min_size=2, max_size=float('inf')):
    """
    Function returning an iterator over found clusters.

    It works by indexing given items into a Vantage Point Tree and then
    querying the tree once per items to find clusters.

    It runs in O(n log n * m), m being the number of matches or leaves to
    traverse in the tree when querying. In practice this is lower than
    O(n^2) but will tend towards it as soon as the number of items becomes too
    large (intuitively, this is because the dimensionality of the queried
    space becomes really high and we hit the curse of high-dimensionality).

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        radius (number): produced clusters' radius.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.

    Yield:
        list: A viable cluster.

    """

    if type(data) is not list:
        data = list(data)

    tree = VPTree(data, distance, selection='spread')

    visited = set()

    for item in data:
        if item in visited:
            continue

        cluster = [neighbor for neighbor, _ in tree.neighbors_in_radius(item, radius)]

        # TODO: with max_size we can use the knn version
        if len(cluster) < min_size or len(cluster) > max_size:
            continue

        visited.update(cluster)

        yield cluster



# =============================================================================
# Fog Graph Utils
# =============================================================================
#
# Miscellaneous utility functions related to graphs.
#
from heapq import nlargest


def component_sizes(g, edge_filter=None):
    seen = set()
    stack = []

    for node in g.nodes:

        if node in seen:
            continue

        c = 0
        stack.append(node)

        while len(stack) != 0:
            n1 = stack.pop()

            if n1 not in seen:
                seen.add(n1)
                c += 1

                for _, n2, edge_attr in g.edges(n1, data=True):
                    if callable(edge_filter) and not edge_filter(n1, n2, edge_attr):
                        continue
                    stack.append(n2)

        yield c


def second_largest_component_size(g, edge_filter=None):
    top = nlargest(2, component_sizes(g, edge_filter))

    if len(top) < 2:
        return None

    return top[1]



# =============================================================================
# Fog Best Matching Cluster Evaluation
# =============================================================================
#
# Implementation of the "best matching F1" evaluation metric.
#
# [References]:
# Yang, Y., Pierce, T., and Carbonell, J. G.  (1998). A study of retrospective
# and on-line event detection. InProc. of ACM-SIGIR, pages 28–36.
#
# Mazoyer, Béatrice, et al. "A french corpus for event detection on twitter."
# Proceedings of The 12th Language Resources and Evaluation Conference. 2020.
# https://www.aclweb.org/anthology/2020.lrec-1.763/
#
from collections import Counter
from typing import Hashable, Iterable, Tuple

from fog.utils import OnlineMean


def best_matching_macro_average(
    truth: Iterable[Iterable[Hashable]],
    predicted: Iterable[Iterable[Hashable]],
    allow_additional_items: bool = False
) -> Tuple[float, float, float]:
    """
    Efficient implementation of the "macro average best matching F1" evaluation
    metric for clusters.

    Note that this metric is not symmetric and will match truth -> predicted.

    Args:
        truth (iterable): the truth clusters.
        predicted (iterable): the predicted clusters.
        allow_additional_items (bool, optional): Whether to allow additional items
            that don't exist in truth clusters to be found in predicted ones. Those
            additional items will then be ignored when computing the metrics instead
            of raising an error when found. Defaults to False.

    Returns:
        tuple of floats: precision, recall and f1 score.

    """

    # NOTE: need to change this if we want to accept lazy iterables
    if len(truth) == 0:
        raise TypeError('truth is empty')

    if len(predicted) == 0:
        raise TypeError('predicted is empty')

    # Indexing truth items
    truth_items = set()

    for cluster in truth:
        for item in cluster:
            truth_items.add(item)

    # Indexing predicted clusters
    index = {}
    predicted_cluster_sizes = {}

    for i, cluster in enumerate(predicted):
        l = 0

        if not cluster:
            raise TypeError('predicted contains an empty cluster')

        for item in cluster:
            if item not in truth_items:
                if not allow_additional_items:
                    raise TypeError('predicted clusters contains items that cannot be found in truth ones')
                else:
                    continue

            if item in index:
                raise TypeError('predicted clusters are fuzzy (i.e. one item can be found in multiple clusters)')

            index[item] = i
            l += 1

        predicted_cluster_sizes[i] = l

    # Aggregating scores
    P = OnlineMean()
    R = OnlineMean()
    F = OnlineMean()

    # Matching truth
    for cluster in truth:
        if not cluster:
            raise TypeError('truth contains an empty cluster')

        # Finding best matching cluster from truth
        candidates = Counter()
        cluster_size = 0

        for item in cluster:
            candidate_cluster_index = index.get(item)

            if candidate_cluster_index is None:
                raise TypeError('truth has items that cannot be found in predicted clusters')

            candidates[candidate_cluster_index] += 1
            cluster_size += 1

        best_f1 = -1.0
        best = None

        # Finding a matching cluster that maximizes F1 score
        for matching_cluster_index, true_positives in candidates.items():
            matching_cluster_size = predicted_cluster_sizes[matching_cluster_index]

            false_positives = matching_cluster_size - true_positives
            false_negatives = cluster_size - true_positives

            precision = true_positives / (true_positives + false_positives)
            recall = true_positives / (true_positives + false_negatives)
            f1 = 2 * precision * recall / (precision + recall)

            if f1 > best_f1:
                best_f1 = f1
                best = (precision, recall, f1)

        assert best is not None

        P.add(best[0])
        R.add(best[1])
        F.add(best[2])

    return (
        float(P),
        float(R),
        float(F)
    )



# =============================================================================
# Fog QuickJoin Clustering
# =============================================================================
#
# Implementation of the Quick Join algorithm that works by recursively
# partitionning the given data with regards to the triangular inequality in
# order to reduce the amount of necessary distance computations.
#
# [Reference]:
# Jacox, Edwin H., et Hanan Samet. « Metric Space Similarity Joins ».
# ACM Transactions on Database Systems 33, no 2 (1 juin 2008): 1‑38.
# https://doi.org/10.1145/1366102.1366104.
#
# Fredriksson K., Braithwaite B. (2013) Quicker Similarity Joins in Metric
# Spaces. In: Brisaboa N., Pedreira O., Zezula P. (eds) Similarity Search and
# Applications. SISAP 2013. Lecture Notes in Computer Science, vol 8199.
# Springer, Berlin, Heidelberg
#
# [Notes]:
# From what I could gather right now, Fredriksson K., Braithwaite B. methods
# to improve the algorithm don't really work with my use-case. For instance,
# the book-keeping of the join_pivots methods takes more time than the
# saved distance computations, even with a eta parameter set to a high value.
# I will need to test examples where the distance is more expensive (e.g.,
# testing with quite tiny strings, the Levensthein distance is not really
# prohibitive right now).
#
# Using a Vantage Point Tree does not yield faster results neither.
#
import dill
import random
from multiprocessing import Pool
from phylactery import VPTree

from fog.clustering.utils import clusters_from_pairs

# TODO: implement the eta parameter


def partition(S, distance, p, radius, rho):
    L = []
    G = []
    Lw = []
    Gw = []

    l = rho - radius
    g = rho + radius

    for item in S:
        d = distance(p, item)

        if d < rho:
            L.append(item)

            if l <= d:
                Lw.append(item)
        else:
            G.append(item)

            if d <= g:
                Gw.append(item)

    return L, G, Lw, Gw


def quickjoin_bruteforce(S1, S2, distance, radius):
    for i in range(len(S1)):
        A = S1[i]

        for j in range(len(S2)):
            B = S2[j]

            if distance(A, B) <= radius:
                yield (A, B)


def quickjoin_self_bruteforce(S, distance, radius):
    n = len(S)

    for i in range(n):
        A = S[i]

        for j in range(i + 1, n):
            B = S[j]

            if distance(A, B) <= radius:
                yield (A, B)


def quickjoin_vptree(S1, S2, distance, radius):

    # Indexing the smallest set
    if len(S1) > len(S2):
        S1, S2 = S2, S1

    if len(S1) == 1:
        B = S1[0]

        for A in S2:
            if distance(A, B) <= radius:
                yield (A, B)
    else:
        tree = VPTree(S1, distance)

        for A in S2:
            for B, d in tree.neighbors_in_radius(A, radius):
                yield (A, B)


def quickjoin_self_vptree(S, distance, radius):

    if len(S) == 1:
        return

    tree = VPTree(S, distance)

    for A in S:
        for B, d in tree.neighbors_in_radius(A, radius):
            if A == B:
                continue

            if id(A) < id(B):
                yield (A, B)


def quickjoin_join_pivots(S1, S2, distance, radius):
    """
    From Fredriksson & Braithwaite. Algorithm 5.

    """

    N1 = len(S1)
    N2 = len(S2)

    k = max(1, min(16, N1 // 8))

    P = [0] * N2 * k

    for i in range(k):
        A = S1[i]

        for j in range(N2):
            B = S2[j]
            d = distance(A, B)
            P[i * N2 + j] = d

            if d <= radius:
                yield (A, B)

    D = [0] * k

    for i in range(k, N1):
        A = S1[i]

        for l in range(k):
            D[l] = distance(S1[l], A)

        for j in range(N2):
            B = S2[j]
            f = False

            for l in range(k):
                if abs(P[l * N2 + j] - D[l]) > radius:
                    f = True
                    break

            if not f and distance(A, B) <= radius:
                yield (A, B)


def select_block_functions(vp_tree):
    if vp_tree:
        return quickjoin_vptree, quickjoin_self_vptree

    return quickjoin_bruteforce, quickjoin_self_bruteforce


def worker(payload):
    distance, radius, S1, S2, vp_tree = payload

    distance = dill.loads(distance)

    BF, SBF = select_block_functions(vp_tree)

    if S2 is None:
        return list(SBF(S1, distance, radius))

    return list(BF(S1, S2, distance, radius))


def quickjoin(data, radius, distance=None, similarity=None, block_size=500,
              min_size=2, max_size=float('inf'),
              mode='connected_components',
              seed=None, processes=1, beta=1.0, vp_tree=False):
    """
    Function returning an iterator over found clusters using the QuickJoin
    algorithm.

    Note that this algorithm returns the same result as pairwise computations
    would.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        distance (callable): The distance function to use. Must be a true
            metric, e.g. the Levenshtein distance.
        similarity (callable): The similarity function to use. Must be a true
            metric, e.g. the Jaccard similarity.
        radius (number): produced clusters' radius.
        block_size (number, optional): block size where the algorithm will
            switch to brute. Defaults to 500.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        mode (string, optional): 'fuzzy_clusters', 'connected_components'.
            Defaults to 'connected_components'.
        seed (number, optional): Seed for RNG. Defaults to None.
        processes (number, optional): Number of processes to use.
            Defaults to 1.
        beta (number, optional): Balancing parameter from Fredriksson &
            Braithwaite. Defaults to no-op 1.0.
        vp_tree (bool, optional): Whether to use Vantage Point Trees to solve
            blocks. Defaults to False.

    Yields:
        list: A viable cluster.

    """

    rng = random.Random(seed)

    if type(data) is not list:
        data = list(data)

    if similarity is not None:
        distance = lambda x, y: -similarity(x, y)
        radius = -radius

    # Iterator recursively partitioning the data set using QuickJoin's method
    def blocks():
        stack = [(data, None)]

        # "Recursivity" through stack
        while len(stack) != 0:
            S1, S2 = stack.pop()

            # QuickJoin procedure
            if S2 is None:

                S = S1
                N = len(S)

                if N <= block_size:
                    yield (S, None)
                    continue

                # Randomly selecting pivots. They must be different
                p1 = rng.randint(0, N - 1)
                p2 = None

                while p2 is None or p1 == p2:
                    p2 = rng.randint(0, N - 1)

                p1 = S[p1]
                p2 = S[p2]

                rho = beta * distance(p1, p2)

                L, G, Lw, Gw = partition(S, distance, p1, radius, rho)

                # Recursion
                stack.append((G, None))
                stack.append((L, None))
                stack.append((Lw, Gw))

            # QuickJoinWin procedure
            else:
                N1 = len(S1)
                N2 = len(S2)
                N = N1 + N2

                if N <= block_size:

                    if N1 == 0 or N2 == 0:
                        continue

                    yield (S1, S2)
                    continue

                # Randomly selecting pivots. They must be different
                p1 = rng.randint(0, N - 1)
                p2 = None

                while p2 is None or p1 == p2:
                    p2 = rng.randint(0, N - 1)

                p1 = S1[p1] if p1 < N1 else S2[p1 - N1]
                p2 = S1[p2] if p2 < N1 else S2[p2 - N1]

                rho = beta * distance(p1, p2)

                L1, G1, Lw1, Gw1 = partition(S1, distance, p1, radius, rho)
                L2, G2, Lw2, Gw2 = partition(S2, distance, p1, radius, rho)

                stack.append((L1, L2))
                stack.append((G1, G2))
                stack.append((Lw1, Gw2))
                stack.append((Gw1, Lw2))

    # Iterator performing bruteforce distance computation over found blocks
    def clustering():

        BF, SBF = select_block_functions(vp_tree)

        for S1, S2 in blocks():
            if S2 is None:
                yield from SBF(S1, distance, radius)
            else:
                yield from BF(S1, S2, distance, radius)

    def clustering_parallel():
        with Pool(processes=processes) as pool:
            pickled_distance = dill.dumps(distance)

            pool_iter = (
                (pickled_distance, radius, S1, S2, vp_tree)
                for S1, S2 in blocks()
            )

            for pairs in pool.imap_unordered(worker, pool_iter):
                yield from pairs

    yield from clusters_from_pairs(
        clustering() if processes == 1 else clustering_parallel(),
        min_size=min_size,
        max_size=max_size,
        mode=mode,
        fuzzy=True
    )



# =============================================================================
# Fog PassJoin Clustering
# =============================================================================
#
# Implementation of the PassJoin algorithm.
#
# [References]:
# Jiang, Yu, Dong Deng, Jiannan Wang, Guoliang Li, et Jianhua Feng.
# « Efficient Parallel Partition-Based Algorithms for Similarity Search and Join
# with Edit Distance Constraints ». In Proceedings of the Joint EDBT/ICDT 2013
# Workshops on - EDBT ’13, 341. Genoa, Italy: ACM Press, 2013.
# https://doi.org/10.1145/2457317.2457382.
#
# Li, Guoliang, Dong Deng, et Jianhua Feng. « A Partition-Based Method for
# String Similarity Joins with Edit-Distance Constraints ». ACM Transactions on
# Database Systems 38, no 2 (1 juin 2013): 1‑33.
# https://doi.org/10.1145/2487259.2487261.
#
# [Urls]:
# http://people.csail.mit.edu/dongdeng/projects/passjoin/index.html
#
from collections import defaultdict
from fog.clustering.utils import clusters_from_pairs


def count_substrings_l(k, s, l):
    """
    Function returning the number of substrings that will be selected by the
    multi-match-aware selection scheme for theshold `k`, for a string of length
    `s` to match strings of length `l`.

    Args:
        k (int): Levenshtein distance threshold.
        s (int): Length of target strings.
        l (int): Length of strings to match.

    Returns:
        int: The number of selected substrings.

    """
    return ((k ** 2 - abs(s - l) ** 2) // 2) + k + 1


def count_keys(k, s):
    """
    Function returning the minimum number of substrings that will be selected by
    the multi-match-aware selection scheme for theshold `k`, for a string of
    length `s` to match any string of relevant length.

    Note that for queries, this number will be higher => range(s - k, s + k + 1).

    Args:
        k (int): Levenshtein distance threshold.
        s (int): Length of target strings.

    Returns:
        int: The number of selected substrings.

    """

    c = 0

    for l in range(s - k, s + 1):
        c += count_substrings_l(k, s, l)

    return c


def sort_key(string):
    """
    Function returning sort value for the strings in PassJoin algorithm. It
    basically order them by decreasing length, then alphabetically as per
    the "4.2 Effective Indexing Strategy" point of the paper.
    """
    return -len(string), string


def partition(k, l):
    """
    Function partitioning a string into k + 1 uneven segments, the shorter
    ones, then the longer ones.

    Args:
        k (int): Levenshtein distance threshold.
        l (int): Length of the string.

    Yields:
        tuple: index, start, length.

    """
    m = k + 1
    a = l // m
    b = a + 1

    large_segments = l - a * m
    small_segments = m - large_segments

    for i in range(small_segments):
        yield (i, i * a, a)

    offset = i * a + a

    for j in range(large_segments):
        yield (i + 1 + j, offset + j * b, b)


def segments(k, string):
    """
    Function yielding a string's k + 1 passjoin segments to index.

    Args:
        k (int): Levenshtein distance threshold.
        string (str): Target string.

    Yields:
        tuple: index, segment.

    """
    for i, start, length in partition(k, len(string)):
        yield (i, string[start:start + length])


def multi_match_aware_interval(k, delta, i, s, pi, li):
    """
    Function returning the interval of relevant substrings to lookup using the
    multi-match-aware substring selection scheme described in the paper.

    Args:
        k (int): Levenshtein distance threshold.
        delta (int): Signed length difference between both considered strings.
        i (int): k + 1 segment index.
        s (int): string length.
        pi (int): k + 1 segment position in target string.
        li (int): k + 1 segment length.

    Returns:
        tuple: start, stop of the interval.

    """
    start1 = pi - i
    end1 = pi + i

    o = k - i

    start2 = pi + delta - o
    end2 = pi + delta + o

    end3 = s - li

    return max(0, start1, start2), min(end1, end2, end3)


def multi_match_aware_substrings(k, string, l, i, pi, li):
    """
    Function yielding relevant substrings to lookup using the multi-match-aware
    substring selection scheme described in the paper.

    Args:
        k (int): Levenshtein distance threshold.
        string (str): Target string.
        l (int): Length of strings to match.
        i (int): k + 1 segment index.
        pi (int): k + 1 segment position in target string.
        li (int): size in characters of the k + 1 segment.

    Yields:
        str: one of the contiguous substrings.

    """
    s = len(string)

    # Note that we need to keep the non-absolute delta for this function
    # to work in both directions, up & down
    delta = s - l

    start, stop = multi_match_aware_interval(k, delta, i, s, pi, li)

    current_substring = None

    for j in range(start, stop + 1):
        substring = string[j:j + li]

        # We skip identical consecutive substrings (to avoid repetition in
        # case of contiguous letter duplication)
        if substring == current_substring:
            continue

        yield substring

        current_substring = substring


def passjoin(data, k, distance, sort=True, min_size=2, max_size=float('inf'),
             mode='connected_components'):
    """
    Function returning an iterator over found clusters using the PassJoin
    algorithm that is able to find every pair of strings having a Levenshtein
    distance less than or equal to a threshold k.

    It works by indexing k + 1 mostly even segments of the considered strings
    into inverted indices splitted by string length and segment index. Then,
    we use another complex but comprehensive string partition scheme to be able
    to produce candidate pairs which are verified using Levenshtein distance.

    It runs in approximately O(nk^2) or O(nk^3) if not sorted. It's therefore
    very scalable with small k values and is undoubtedly better than the naive
    quadratic approach.

    Args:
        data (iterable): Arbitrary iterable containing data points to gather
            into clusters. Will be fully consumed.
        k (number): Levenshtein distance threshold.
        distance (callable): Function tasked to compute the Levenshtein distance
            between two points of data.
        sort (boolean, optional): whether to sort the data beforehand. Defaults
            to True.
        min_size (number, optional): minimum number of items in a cluster for
            it to be considered viable. Defaults to 2.
        max_size (number, optional): maximum number of items in a cluster for
            it to be considered viable. Defaults to infinity.
        mode (string, optional): 'fuzzy_clusters', 'connected_components'.
            Defaults to 'connected_components'.

    Yields:
        list: A viable cluster.

    """

    if sort:

        # NOTE: sorting in reverse as per "4.2 Effective Indexing Strategy"
        data = sorted(data, key=sort_key)

    # TODO: when keys lengths are <= k some pairs are tested more than once!
    # TODO: test exact match cases
    # TODO: empty string should probably be handled in its own way?

    def clustering():
        L = defaultdict(lambda: defaultdict(list))
        Ll = None

        for A in data:
            s = len(A)

            # Attempting to match the string
            for l in range(s if sort else max(0, s - k), s + k + 1):
                Ll = L.get(l)

                # Index is empty for this length, no need to enquire further
                if Ll is None:
                    continue

                for i, start, length in partition(k, l):
                    for substring in multi_match_aware_substrings(k, A, l, i, start, length):
                        candidates = Ll.get((i, substring))

                        if candidates is None:
                            continue

                        for B in candidates:

                            # NOTE: first condition is here not to compute Levenshtein
                            # distance for tiny strings
                            # NOTE: a pair may arise more than once here
                            # It's taken care of later but I feel we can do better...
                            # What's more, we should also record non-matches even if the
                            # case of testing twice is naturally rarer
                            if (s <= k and l <= k) or distance(A, B) <= k:
                                yield (A, B)

            # Indexing the string
            # NOTE: it's possible to cleanup some memory when working on sorted strings
            Ll = L[s]

            for key in segments(k, A):
                Ll[key].append(A)

    yield from clusters_from_pairs(
        clustering(),
        min_size=min_size,
        max_size=max_size,
        mode=mode
    )



# =============================================================================
# Fog Graph Projections
# =============================================================================
#
# Miscellaneous functions related to bipartite to monopartite projections.
#
import math
import networkx as nx

from fog.metrics.cosine import sparse_dot_product
from fog.metrics.utils import intersection_size

MONOPARTITE_PROJECTION_METRICS = ('cosine', 'jaccard', 'overlap', 'dice', 'binary_cosine')


def compute_metric(metric, vector1, vector2, norm1, norm2):
    if metric == 'cosine':
        w = sparse_dot_product(vector1, vector2)

        if w == 0:
            return 0

        return w / (norm1 * norm2)

    w = intersection_size(vector1, vector2)

    if w == 0:
        return 0

    if metric == 'jaccard':
        return w / (norm1 + norm2 - w)

    if metric == 'dice':
        return (2 * w) / (norm1 + norm2)

    if metric == 'overlap':
        return w / min(norm1, norm2)

    if metric == 'binary_cosine':
        return w / math.sqrt(norm1 * norm2)

    return w


# TODO: use passjoin prefix filtering as optimization scheme
def monopartite_projection(bipartite, project, part='bipartite', weight='weight',
                           metric=None, threshold=None, use_topology=True,
                           bipartition_check=True):
    """
    Function computing a monopartite projection of the given bipartite graph.
    This projection can be basic and create a weighted edge each time two nodes
    in target partition share a common neighbor. Or it can be weighted and
    filtered using a similarity metric such as Jaccard or cosine similarity,
    for instance.

    Args:
        bipartite (nx.Graph): Target bipartite graph.
        project (str): Name of the partition to project.
        part (str, optional): Name of the node attribute on which the
            graph partition is built e.g. "color" or "type" etc.
            Defaults to "bipartite".
        weight (str, optional): Name of the weight edge attribute.
            Defaults to "weight".
        metric (str, optional): Metric to use. If `None`, the basic projection
            will be returned. Also accepts `jaccard`, `overlap`, `dice`,
            `cosine` or `binary_cosine`. Defaults to None.
        threshold (float, optional): Optional similarity threshold under which
            edges won't be added to the monopartite projection.
            Defaults to None.
        use_topology (bool, optional): Whether to use the bipartite graph's
            topology to attempt a subquadratic time projection. Intuitively,
            this works by not computing similarities of all pairs of nodes but
            only of pairs of nodes that share at least a common neighbor.
            It generally works better than the quadratic approach but can
            sometimes hurt your performance by losing time on graph traversals
            when your graph is very dense.
        bipartition_check (bool, optional): This function will start by checking
            whether your graph is bipartite because it can get stuck in an
            infinite loop if given graph is not truly bipartite. Be sure to
            disable this kwarg if you know beforehand that your graph is
            bipartite and for better performance.

    Returns:
        nx.Graph: The monopartite projection.

    """

    if metric is not None and metric not in MONOPARTITE_PROJECTION_METRICS:
        raise TypeError('fog.graph.monopartite_projection: unsupported metric "%s"' % metric)

    if bipartition_check:
        parts = set()

        for n1, n2 in bipartite.edges:
            p1 = bipartite.nodes[n1][part]
            p2 = bipartite.nodes[n2][part]

            parts.add(p1)
            parts.add(p2)

            if p1 == p2 or len(parts) > 2:
                raise TypeError('fog.graph.monopartite_projection: given graph is not truly bipartite!')

    monopartite = nx.Graph()

    # Initialising monopartite graph
    for node, attr in bipartite.nodes(data=True):
        if attr[part] != project:
            continue

        monopartite.add_node(node, **attr)

    # Accumulating vectors & norms
    # TODO: we could save up some memory by relying on nx's graph directly
    # But this would make code a bit more complex and won't ease up adapting
    # the function to a sparse vector input
    vectors = {}

    for node in monopartite.nodes:
        s = 0
        neighbors = {} if metric == 'cosine' else set()

        for _, neighbor, w in bipartite.edges(node, data=weight, default=1):
            if metric == 'cosine':
                s += w * w
                neighbors[neighbor] = w
            else:
                s += 1
                neighbors.add(neighbor)

        if s > 0:
            if metric == 'cosine':
                vectors[node] = (math.sqrt(s), neighbors)
            else:
                vectors[node] = (s, neighbors)

    # Version relying on graph traversals
    if use_topology:

        for n1, (norm1, vector1) in vectors.items():
            for np in bipartite.neighbors(n1):
                for n2 in bipartite.neighbors(np):

                    # We test only (n1, n2) and not (n2, n1) which is pointless
                    # since our metrics are symmetric. This is the basically
                    # the same as cutting edges from n1 to np after you
                    # traversed them.
                    # We also avoid (n1, n1) since n1 is obv. a neighbor of np.
                    if n1 >= n2:
                        continue

                    # NOTE: if a threshold is given, this condition does
                    # not work as expected since edges are not added to
                    # the monopartite graph if they fall below the threshold
                    # It is possible to store a set of failed edges but
                    # it can cost a lot of memory and requires an additional
                    # lookup. Furthermore the probability to test several times
                    # the same failed edge is usually far lower than for a
                    # valid one, which makes hard to know if spending memory
                    # and lookups is better than actually computing the
                    # similarity metric more times than necessary.
                    if monopartite.has_edge(n1, n2):
                        continue

                    norm2, vector2 = vectors[n2]

                    # NOTE: at this point, both norms should be > 0
                    w = compute_metric(metric, vector1, vector2, norm1, norm2)

                    if w == 0:
                        continue

                    if threshold is not None and w < threshold:
                        continue

                    monopartite.add_edge(n1, n2, **{weight: w})

        return monopartite

    # Quadratic version
    nodes = list(vectors.items())
    l = len(nodes)

    for i, (n1, (norm1, vector1)) in enumerate(nodes):
        for j in range(i + 1, l):
            (n2, (norm2, vector2)) = nodes[j]

            # NOTE: at this point, both norms should be > 0
            w = compute_metric(metric, vector1, vector2, norm1, norm2)

            if w == 0:
                continue

            if threshold is None or w >= threshold:
                monopartite.add_edge(n1, n2, **{weight: w})

    return monopartite



# =============================================================================
# Fog MinHash Clustering
# =============================================================================
#
# Clustering algorithm leveraging MinHash LSH to produce suitable clusters.
#
# [Url]:
# http://infolab.stanford.edu/~ullman/mmds/ch3.pdf
#
from collections import defaultdict

from fog.clustering.utils import clusters_from_buckets
from fog.lsh.minhash import MinHash
from fog.metrics.jaccard import jaccard_similarity


# TODO:
#   * Parallelize
#   * possibility to hash the band key
#   * note that we allow uneven bands for fine grained results
#   * double_check with minhash or jaccard or sub similarity even
#   * superminhash to generate signature faster (better for large docs)
#   * cheap_hashes
#   * possibility to use one dict per band + sum the integers

# TODO: compute similarities online + edge list -> connected components
# TODO: keep edge list in set of tuples with sorted comp not to compute
# n times the same similarity
# TODO: use pairs


def match_probability(h, bands, similarity):
    """
    Function returning the probability two pairs will match given a number
    of a signature's integers, the number of bands dividing the signature
    matrix and the desired similarity.

    Args:
        h (int): Number of integers in the minhash signature.
        bands (int): Number of bands dividing the signature matrix.
        similarity (float): Desired Jaccard similarity.

    Returns:
        float: The match probability.

    """
    return 1.0 - (1.0 - similarity ** (h / bands)) ** bands


def similarity_threshold(h, bands):
    """
    Function returning the Jaccard similarity threshold for minhash signature
    composed of h integers and a signature matrix divided in n bands.

    Args:
        h (int): Number of integers in the minhash signature.
        bands (int): Number of bands dividing the signature matrix.

    Returns:
        float: The Jaccard similarity threshold.

    """
    return (1.0 / bands) ** (1 / (h / bands))


def guess_bands(h, threshold):
    """
    Function used to iteratively guess the optimal number of bands needed to
    divide a minhash signature matrix in order to find pairs having a
    Jaccard similarity over the given threshold.

    Args:
        h (int): Number of integers in the minhash signature.
        threshold (float): Jaccard similarity threshold.

    Returns:
        int: The optimal number of bands.

    """

    bands = 1

    while bands <= h:
        t = similarity_threshold(h, bands)

        if t <= threshold:
            break

        bands += 1

    return bands


def minhash(data, h=256, key=None, radius=0.8, bands=None, use_numpy=False,
            seed=None):
    """
    Function returning an iterator over clusters found using the minhash
    clustering method.

    The idea is to compute minhash signatures for every item and divide the
    resulting signature matrix in bands of n rows so that if two items share
    the exact same rows in a band, they are likely to be similar.

    It runs in O(nh), n being the number of items, h the number of integers to
    use as minhash signature. Note that since usually h << n, it practically
    runs in O(n).

    Args:
        data (iterable): Items to cluster.
        h (int, optional): Number of integers to use as the minhash signature.
            Defaults to 256.
        key (callable, optional): Function returning an item's key.
        radius (float, optional): Radius over which a pair of items is deemed
            similar. Defaults to 0.8.
        bands (int, optional): By defaults, the function will attempt to guess
            the optimal number of bands to use to divide the signature matrix
            using given radius. Set this argument if you want to set the
            number of bands by yourself.
        use_numpy (bool, optional): whether to use numpy to speed up minhash
            signatures computations. Defaults to False.
        seed (int, optional): rng seed.

    """

    if bands is None:
        bands = guess_bands(h, radius)

    rows = h // bands
    h_upper_bound = bands * rows

    mh = MinHash(h, use_numpy=use_numpy, seed=seed)

    buckets = defaultdict(list)

    for item in data:
        k = item

        if key is not None:
            k = key(item)

        signature = mh.create_signature(k)

        for band in range(0, h_upper_bound, rows):
            band_key = (band, '%'.join(str(n) for n in signature[band:band + rows]))
            buckets[band_key].append(item)

    def double_check(A, B):
        if key is not None:
            return jaccard_similarity(key(A), key(B)) >= radius

        return jaccard_similarity(A, B) >= radius

    yield from clusters_from_buckets(
        buckets.values(),
        mode='connected_components',
        similarity=double_check
    )



# =============================================================================
# Fog Clustering Utilities
# =============================================================================
#
# Miscellaneous redundant functions used by clustering routines.
#
from collections import defaultdict
import math


def make_similarity_function(similarity=None, distance=None, radius=None):
    """
    Function creating a similarity function returning True if the compared
    items are similar from a variety of functions & parameters.

    Basically, if a distance function is given, it will be inverted and if
    a radius is given, the returned function will return whether the
    distance/similarity between two items is under/over it.

    Args:
        similarity (callable, optional): Similarity function.
        distance (callable, optional): Distance function.
        radius (number, optional): Radius.

    Returns:
        function: A similarity function with signature (A, B) -> bool

    """
    if similarity is None and distance is None:
        raise TypeError('fog.clustering: need at least a similarity or distance function.')

    if radius is not None:
        if similarity:
            return lambda A, B: similarity(A, B) >= radius
        else:
            return lambda A, B: distance(A, B) <= radius
    else:
        if similarity:
            return similarity
        else:
            return lambda A, B: not distance(A, B)


def clusters_from_pairs(pairs, min_size=2, max_size=float('inf'),
                        mode='connected_components', fuzzy=False):
    """
    Function consuming an iterator of similar pairs and merging them
    according to the desired strategy to yield valid clusters.

    Args:
        pairs (iterable): Similar pairs.
        min_size (int, optional): Minimum size of clusters, defaults to 2.
        max_size (int, optional): Maximum size of clusters, defaults to
            infinity.
        mode (string, optional): 'fuzzy_clusters', 'connected_components'.
            Defaults to 'connected_components'.
        fuzzy (bool, optional): whether a same pair can arrive twice or not.
            Defaults to False.

    Yields:
        list: A viable cluster.

    """

    container = set if fuzzy else list

    if fuzzy:
        def add(c, x):
            c.add(x)
    else:
        def add(c, x):
            c.append(x)

    # TODO: later, we'll be able to use SparseSet rather than the graph
    # TODO: in the connected component case, we don't need mutual links
    graph = defaultdict(container)

    for A, B in pairs:
        add(graph[A], B)
        add(graph[B], A)

    if mode == 'fuzzy_clusters':
        visited = set()

        for item, neighbors in graph.items():
            if item in visited:
                continue

            if len(neighbors) + 1 < min_size:
                continue
            if len(neighbors) + 1 > max_size:
                continue

            visited.update(neighbors)

            cluster = [item] + (list(neighbors) if fuzzy else neighbors)
            yield cluster

    elif mode == 'connected_components':
        visited = set()
        stack = []

        for item, neighbors in graph.items():
            if item in visited:
                continue

            visited.add(item)

            cluster = [item]

            stack.extend(neighbors)

            while len(stack) != 0:
                neighbor = stack.pop()

                if neighbor in visited:
                    continue

                cluster.append(neighbor)
                visited.add(neighbor)

                stack.extend(graph[neighbor])

            if len(cluster) >= min_size and len(cluster) <= max_size:
                yield cluster

    else:
        raise TypeError('fog.clustering: unknown mode "%s"' % mode)


def pairs_from_buckets(buckets):
    """
    Function yielding similarity pairs from buckets.

    """
    for bucket in buckets:
        n = len(bucket)

        for i in range(n):
            A = bucket[i]

            for j in range(i + 1, n):
                B = bucket[j]

                yield (A, B)


def clusters_from_buckets(buckets, min_size=2, max_size=float('inf'),
                          mode='connected_components', similarity=None,
                          fuzzy=False):
    """
    Function merging buckets into fuzzy clusters. Each bucket will create
    relations in an undirected graph that is later solved to compose clusters.

    Args:
        buckets (iterable): Buckets to merge.
        min_size (int, optional): Minimum size of clusters, defaults to 2.
        max_size (int, optional): Maximum size of clusters, defaults to
            infinity.
        mode (string, optional): 'fuzzy_clusters', 'connected_components' or
            'leader'. Defaults to 'connected_components'.
        similarity (callable, optional)= similarity function to use to validate
            matches from buckets.
        fuzzy (bool, optional): whether a same pair can arrive twice or not.
            Defaults to False.

    Yields:
        list: A viable cluster.

    """

    pairs = (
        pair
        for pair
        in pairs_from_buckets(buckets)
        if similarity is None or similarity(pair[0], pair[1])
    )

    yield from clusters_from_pairs(
        pairs,
        min_size=min_size,
        max_size=max_size,
        mode=mode,
        fuzzy=fuzzy
    )


def upper_triangular_matrix_chunk_iter(data, chunk_size):
    """
    Function returning an iterator over chunks of an upper triangular matrix.
    It's a useful utility to parallelize pairwise distance computations, for
    instance.

    Args:
        data (iterable): The matrix's data.
        chunk_size (int): Size of the chunks to yield.

    Yields:
        tuple of slices: A matrix's chunk.

    """
    n = len(data)
    c = math.ceil(n / chunk_size)

    for j in range(c):
        j_offset = j * chunk_size
        j_limit = j_offset + chunk_size

        for i in range(0, min(j + 1, c)):
            i_offset = i * chunk_size
            i_limit = i_offset + chunk_size

            yield (
                data[i_offset:i_limit],
                data[j_offset:j_limit] if i_offset != j_offset else [],
                i_offset,
                j_offset
            )



# =============================================================================
# Fog Overlap Coefficient
# =============================================================================
#
# Functions computing the overlap coefficient.
#
# [Urls]:
# https://en.wikipedia.org/wiki/Overlap_coefficient
#
from fog.metrics.utils import ACCEPTABLE_TYPES


def overlap_coefficient(A, B):
    """
    Function computing the overlap coefficient of the given sets, i.e. the size
    of their intersection divided by the size of the smallest set.

    Runs in O(n), n being the size of the smallest set.

    Args:
        A (iterable): First sequence.
        B (iterable): Second sequence.

    Returns:
        float: overlap coefficient between A & B.

    """
    if A is B:
        return 1.0

    if not isinstance(A, ACCEPTABLE_TYPES):
        A = set(A)

    if not isinstance(B, ACCEPTABLE_TYPES):
        B = set(B)

    if len(A) == 0 or len(B) == 0:
        return 0.0

    # Swapping to iterate over smaller set and minimize lookups
    if len(A) > len(B):
        A, B = B, A

    # Counting intersection
    I = 0

    for v in A:
        if v in B:
            I += 1

    return I / min(len(A), len(B))



# =============================================================================
# Fog Cluster CLI Action
# =============================================================================
#
# Logic of the cluster CLI action enabling the user to cluster the content of
# a CSV file.
#
from collections import Counter
from datetime import datetime
from timeit import default_timer as timer

from fog.cli.utils import custom_reader
from fog.cli.reporting import RENDERERS
from fog.clustering import (
    key_collision
)
from fog.key import fingerprint

CLUSTERING_ROUTINES = {
    'fingerprint_collision': {
        'name': 'Fingerpint collision',
        'fn': key_collision,
        'kwargs': {
            'key': fingerprint
        }
    }
}


def cluster_action(namespace):
    routine = CLUSTERING_ROUTINES[namespace.algorithm]

    lines = 0

    headers, position, reader = custom_reader(namespace.file, namespace.column)

    values = Counter()

    for line in reader:
        lines += 1
        values[line[position]] += 1

    start = timer()
    clusters = routine['fn'](values.keys(), **routine['kwargs'])
    end = timer() - start

    meta = {
        'algorithm': namespace.algorithm,
        'took': end,
        'date': datetime.now(),
        'lines': lines
    }

    renderer = RENDERERS[namespace.type]

    renderer(namespace.output, meta, values, list(clusters))



# =============================================================================
# Fog Cluster TOML Reporting Utilities
# =============================================================================
#
# Function used to render a clustering result into a TOML file.
#


def escape_string(string):
    return string.replace('"', '\\"').replace('\n', '\\n')


def print_toml_report(file, meta, values, clusters):

    def p(*args):
        print(*args, file=file)

    p('[info]')
    p('date = %s' % meta['date'].isoformat().split('.')[0])
    p('algorithm = "%s"' % meta['algorithm'])
    p()
    p('[stats]')
    p('lines = %i' % meta['lines'])
    p('nb_distinct_values = %i' % len(values))
    p('nb_clusters = %i' % len(clusters))
    p('took = %2f' % meta['took'])
    p()

    for i, cluster in enumerate(clusters):
        p('[[cluster]]')
        p('id = %i' % i)
        p('nb_values = %i' % len(cluster))
        p('nb_rows = %i' % sum(values[v] for v in cluster))

        # Sorting by affected rows
        sorted_values = sorted(cluster, key=lambda v: values[v], reverse=True)
        max_length = len(max(cluster, key=len))

        p('harmonized = "%s"' % escape_string(sorted_values[0]))
        p('values =  [')
        for value in sorted_values:
            p('  [["%s"],%s [%i]],' % (
                escape_string(value),
                ' ' * (max_length - len(value)),
                values[value]
            ))

        p(']')

        p('harmonize = false')
        p('')



# =============================================================================
# Fog Cluster HTML Reporting Utilities
# =============================================================================
#
# Function used to render a clustering result into a HTML file.
#
import datetime


def print_html_report(file, meta, values, clusters):

    def p(*args):
        print(*args, file=file)

    p('<html>')
    p('  <head>')
    p('    <title>Clustering result</title>')
    p('    <style>')
    p('      td {padding-right: 10px};')
    p('    </style>')
    p('  </head>')
    p('  <body>')
    p('    <h1>Clustering result</h1>')
    p('    <div>')
    p('      <h2>Information:</h2>')
    p('      <table>')
    p('        <tr>')
    p('          <td>')
    p('            <b>Date</b>')
    p('          </td>')
    p('          <td>')
    p('            %s' % meta['date'].isoformat().split('T')[0])
    p('          </td>')
    p('        </tr>')
    p('        <tr>')
    p('          <td>')
    p('            <b>Algorithm used</b>')
    p('          </td>')
    p('          <td>')
    p('            %s' % meta['algorithm'])
    p('          </td>')
    p('        </tr>')
    p('      </table>')
    p('    </div>')
    p('    <div>')
    p('      <h2>Statistics:</h2>')
    p('      <table>')
    p('        <tr>')
    p('          <td>')
    p('            <b>Lines</b>')
    p('          </td>')
    p('          <td>')
    p('            {:,}'.format(meta['lines']))
    p('          </td>')
    p('        </tr>')
    p('        <tr>')
    p('          <td>')
    p('            <b>Distinct values</b>')
    p('          </td>')
    p('          <td>')
    p('            {:,}'.format(len(values)))
    p('          </td>')
    p('        </tr>')
    p('        <tr>')
    p('          <td>')
    p('            <b>Found clusters</b>')
    p('          </td>')
    p('          <td>')
    p('            {:,}'.format(len(clusters)))
    p('          </td>')
    p('        </tr>')
    p('        <tr>')
    p('          <td>')
    p('            <b>Took</b>')
    p('          </td>')
    p('          <td>')
    p('            %s' % datetime.timedelta(microseconds=meta['took']))
    p('          </td>')
    p('        </tr>')
    p('      </table>')
    p('    </div>')
    p('    <div>')
    p('      <h2>Clusters:</h2>')

    for i, cluster in enumerate(clusters):
        p('      <p>')
        p('        Cluster n°%i (%i unique values, %i affected rows):' % (i + 1, len(cluster), sum(values[v] for v in cluster)))
        p('      <ul>')

        # Sorting by affected rows
        sorted_values = sorted(cluster, key=lambda v: values[v], reverse=True)

        for v in sorted_values:
            p('        <li>')
            p('          %s' % v)
            p('        </li>')

        p('      </ul>')
        p('      </p>')

    p('    </div>')
    p('  </body>')
    p('<html>')



from fog.tokenizers.fingerprint import (
    FingerprintTokenizer,
    NgramsFingerprintTokenizer
)
from fog.tokenizers.ngrams import ngrams, bigrams, trigrams, quadrigrams
from fog.tokenizers.words import WordTokenizer

fingerprint_tokenizer = FingerprintTokenizer()
ngrams_fingerprint_tokenizer = NgramsFingerprintTokenizer()
words_tokenizer = WordTokenizer()
